from __future__ import annotations

from pathlib import Path
from typing import Any

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QWidget, QFormLayout, QPushButton, QFileDialog, QLineEdit, QHBoxLayout, QVBoxLayout, \
    QGridLayout, QLabel
from reactives.instance.method import reactive_method

from betty.app import App
from betty.extension import Gramps
from betty.extension.gramps.config import FamilyTreeConfiguration
from betty.gui import BettyWindow, mark_valid, mark_invalid
from betty.gui.error import catch_exceptions
from betty.gui.locale import LocalizedWidget
from betty.gui.text import Text
from betty.serde.error import SerdeError


class _FamilyTrees(LocalizedWidget):
    def __init__(self, app: App, *args: Any, **kwargs: Any):
        super().__init__(app, *args, **kwargs)

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._family_trees_widget = QWidget()
        self._family_trees_layout = QGridLayout()
        self._family_trees_remove_buttons: list[QPushButton] = []

        self._build_family_trees()

        self._add_family_tree_button = QPushButton()
        self._add_family_tree_button.released.connect(self._add_family_tree)
        self._layout.addWidget(self._add_family_tree_button, 1)

    @reactive_method(on_trigger_call=True)
    def _build_family_trees(self) -> None:
        self._family_trees_widget = QWidget()
        self._family_trees_layout = QGridLayout()
        self._family_trees_remove_buttons = []
        self._family_trees_widget.setLayout(self._family_trees_layout)
        self._layout.addWidget(self._family_trees_widget)

        for i, family_tree in enumerate(self._app.extensions[Gramps].configuration.family_trees):
            def _remove_family_tree() -> None:
                del self._app.extensions[Gramps].configuration.family_trees[i]
            self._family_trees_layout.addWidget(Text(str(family_tree.file_path)), i, 0)
            self._family_trees_remove_buttons.insert(i, QPushButton())
            self._family_trees_remove_buttons[i].released.connect(_remove_family_tree)
            self._family_trees_layout.addWidget(self._family_trees_remove_buttons[i], i, 1)
        self._layout.insertWidget(0, self._family_trees_widget, alignment=Qt.AlignmentFlag.AlignTop)

    def _do_set_translatables(self) -> None:
        self._add_family_tree_button.setText(self._app.localizer._('Add a family tree'))
        for button in self._family_trees_remove_buttons:
            button.setText(self._app.localizer._('Remove'))

    def _add_family_tree(self) -> None:
        window = _AddFamilyTreeWindow(self._app, self)
        window.show()


class _GrampsGuiWidget(LocalizedWidget):
    def __init__(self, app: App, *args: Any, **kwargs: Any):
        super().__init__(app, *args, **kwargs)
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._family_trees = _FamilyTrees(self._app)
        self._layout.addWidget(self._family_trees)


class _AddFamilyTreeWindow(BettyWindow):
    window_width = 500
    window_height = 100

    def __init__(self, app: App, *args: Any, **kwargs: Any):
        super().__init__(app, *args, **kwargs)
        self._family_tree = FamilyTreeConfiguration()

        self._layout = QFormLayout()

        self._widget = QWidget()
        self._widget.setLayout(self._layout)

        self.setCentralWidget(self._widget)

        def _update_configuration_file_path(file_path: str) -> None:
            if not file_path:
                self._save_and_close.setDisabled(True)
                return
            try:
                self._family_tree.file_path = Path(file_path)
                mark_valid(self._file_path)
                self._save_and_close.setDisabled(False)
            except SerdeError as e:
                mark_invalid(self._file_path, str(e))
                self._save_and_close.setDisabled(True)
        self._file_path = QLineEdit()
        self._file_path.textChanged.connect(_update_configuration_file_path)
        file_path_layout = QHBoxLayout()
        file_path_layout.addWidget(self._file_path)

        @catch_exceptions
        def find_family_tree_file_path() -> None:
            found_family_tree_file_path, __ = QFileDialog.getOpenFileName(
                self._widget,
                self._app.localizer._('Load the family tree from...'),
                directory=self._file_path.text(),
            )
            if '' != found_family_tree_file_path:
                self._file_path.setText(found_family_tree_file_path)
        self._file_path_find = QPushButton('...')
        self._file_path_find.released.connect(find_family_tree_file_path)
        file_path_layout.addWidget(self._file_path_find)
        self._file_path_label = QLabel()
        self._layout.addRow(self._file_path_label, file_path_layout)

        buttons_layout = QHBoxLayout()
        self._layout.addRow(buttons_layout)

        @catch_exceptions
        def save_and_close_family_tree() -> None:
            self._app.extensions[Gramps].configuration.family_trees.append(self._family_tree)
            self.close()
        self._save_and_close = QPushButton()
        self._save_and_close.setDisabled(True)
        self._save_and_close.released.connect(save_and_close_family_tree)
        buttons_layout.addWidget(self._save_and_close)

        self._cancel = QPushButton()
        self._cancel.released.connect(self.close)
        buttons_layout.addWidget(self._cancel)

    def _do_set_translatables(self) -> None:
        self._file_path_label.setText(self._app.localizer._('File path'))
        self._save_and_close.setText(self._app.localizer._('Save and close'))
        self._cancel.setText(self._app.localizer._('Cancel'))

    @property
    def title(self) -> str:
        return self._app.localizer._('Add a family tree')



from typing import Any, TypeAlias

DictState: TypeAlias = dict[str, Any]

SlotsState: TypeAlias = dict[str, Any]

State: TypeAlias = tuple[DictState, SlotsState]


class Pickleable:
    def __getstate__(self) -> State:
        return {}, {}

    def __setstate__(self, state: State):
        dict_state, slots_state = state
        self.__dict__.update(**dict_state)
        for slot_name, slot_value in slots_state.items():
            setattr(self, slot_name, slot_value)



import logging
import os
import subprocess as stdsubprocess
from asyncio import subprocess
from textwrap import indent
from typing import Sequence, Any, Callable, Awaitable, ParamSpec


P = ParamSpec('P')


async def run_exec(runnee: Sequence[str], **kwargs: Any) -> subprocess.Process:
    return await _run(
        subprocess.create_subprocess_exec,
        runnee,
        runnee[0],
        *runnee[1:],
        **kwargs,
    )


async def run_shell(runnee: Sequence[str], **kwargs: Any) -> subprocess.Process:
    return await _run(
        subprocess.create_subprocess_shell,
        runnee,
        ' '.join(runnee),
        **kwargs,
    )


async def _run(
    runner: Callable[P, Awaitable[subprocess.Process]],
    runnee: Sequence[str],
    *args: P.args,
    **kwargs: P.kwargs,
) -> subprocess.Process:
    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.PIPE
    process = await runner(*args, **kwargs)
    await process.wait()

    if process.returncode == 0:
        return process

    stdout = process.stdout
    stdout_str = '' if stdout is None else '\n'.join((await stdout.read()).decode().split(os.linesep))
    stderr = process.stderr
    stderr_str = '' if stderr is None else '\n'.join((await stderr.read()).decode().split(os.linesep))

    error = stdsubprocess.CalledProcessError(
        process.returncode,  # type: ignore[arg-type]
        ' '.join(runnee),
        stdout_str,
        stderr_str,
    )
    logging.getLogger().warning(
        f'{str(error)}\nSTDOUT:\n{indent(stdout_str, "    ")}\nSTDERR:{indent(stderr_str, "   ")}',
    )
    raise error



from __future__ import annotations

import contextlib
import logging
import threading
import webbrowser
from http.server import SimpleHTTPRequestHandler, HTTPServer
from io import StringIO
from types import TracebackType
from typing import Sequence

from aiofiles.os import makedirs

from betty.app import App
from betty.asyncio import sync
from betty.error import UserFacingError
from betty.locale import Str, Localizer

DEFAULT_PORT = 8000


class ServerNotStartedError(RuntimeError):
    pass


class NoPublicUrlBecauseServerNotStartedError(ServerNotStartedError):
    def __init__(self):
        super().__init__('Cannot get the public URL for a server that has not started yet.')


class OsError(UserFacingError, OSError):
    pass


class Server:
    """
    Provide a development web server.
    """

    def __init__(self, localizer: Localizer):
        self._localizer = localizer

    @classmethod
    def name(cls) -> str:
        return f'{cls.__module__}.{cls.__name__}'

    @classmethod
    def label(cls) -> Str:
        raise NotImplementedError(repr(cls))

    async def start(self) -> None:
        """
        Starts the server.
        """
        pass

    async def show(self) -> None:
        """
        Shows the served site to the user.
        """
        logging.getLogger().info(self._localizer._('Serving your site at {url}...').format(
            url=self.public_url,
        ))
        webbrowser.open_new_tab(self.public_url)

    async def stop(self) -> None:
        """
        Stops the server.
        """
        pass

    @property
    def public_url(self) -> str:
        raise NotImplementedError(repr(self))

    async def __aenter__(self) -> Server:
        await self.start()
        return self

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        await self.stop()


class AppServer(Server):
    def __init__(self, app: App) -> None:
        super().__init__(app.localizer)
        self._app = app

    @staticmethod
    def get(app: App) -> AppServer:
        for server in app.servers.values():
            if isinstance(server, AppServer):
                return server
        raise RuntimeError(f'Cannot find a project server. This must never happen, because {BuiltinServer} should be the fallback.')

    async def start(self) -> None:
        await makedirs(self._app.project.configuration.www_directory_path, exist_ok=True)
        await super().start()


class ServerProvider:
    @property
    def servers(self) -> Sequence[Server]:
        raise NotImplementedError(repr(self))


class _BuiltinServerRequestHandler(SimpleHTTPRequestHandler):
    def end_headers(self) -> None:
        self.send_header('Cache-Control', 'no-cache')
        super().end_headers()


class BuiltinServer(AppServer):
    def __init__(self, app: App) -> None:
        super().__init__(app)
        self._http_server: HTTPServer | None = None
        self._port: int | None = None
        self._thread: threading.Thread | None = None

    @classmethod
    def label(cls) -> Str:
        return Str._('Python built-in')

    async def start(self) -> None:
        await super().start()
        logging.getLogger().info(self._localizer._("Starting Python's built-in web server..."))
        for self._port in range(DEFAULT_PORT, 65535):
            with contextlib.suppress(OSError):
                self._http_server = HTTPServer(
                    ('', self._port),
                    lambda request, client_address, server: _BuiltinServerRequestHandler(
                        request,
                        client_address,
                        server,
                        directory=str(self._app.project.configuration.www_directory_path),
                    ),
                )
                break
        if self._http_server is None:
            raise OsError(Str._('Cannot find an available port to bind the web server to.'))
        self._thread = threading.Thread(target=self._serve)
        self._thread.start()

    @property
    def public_url(self) -> str:
        if self._port is not None:
            return f'http://localhost:{self._port}'
        raise NoPublicUrlBecauseServerNotStartedError()

    @sync
    async def _serve(self) -> None:
        with contextlib.redirect_stderr(StringIO()):
            assert self._http_server
            self._http_server.serve_forever()

    async def stop(self) -> None:
        await super().stop()
        if self._http_server is not None:
            self._http_server.shutdown()
            self._http_server.server_close()
        if self._thread is not None:
            self._thread.join()



from __future__ import annotations

from contextlib import AsyncExitStack

from geopy import Point

from betty import load, generate
from betty.app import App
from betty.app.extension import Extension
from betty.load import Loader
from betty.locale import Date, DateRange, Str
from betty.model import Entity
from betty.model.ancestry import Place, PlaceName, Person, Presence, Subject, PersonName, Link, Source, Citation, Event, \
    Enclosure
from betty.model.event_type import Marriage, Birth, Death
from betty.project import LocaleConfiguration, ExtensionConfiguration, EntityReference
from betty.serve import Server, AppServer, NoPublicUrlBecauseServerNotStartedError


class _Demo(Extension, Loader):
    @classmethod
    def depends_on(cls) -> set[type[Extension]]:
        from betty.extension import CottonCandy, HttpApiDoc, Maps, Trees, Wikipedia

        return {CottonCandy, HttpApiDoc, Maps, Trees, Wikipedia}

    def _load(self, *entities: Entity) -> None:
        self._app.project.ancestry.add(*entities)

    async def load(self) -> None:
        from betty.extension import CottonCandy

        netherlands = Place(
            id='betty-demo-netherlands',
            names=[
                PlaceName(name='Netherlands'),
                PlaceName(
                    name='Nederland',
                    locale='nl',
                ),
                PlaceName(
                    name='Нідерланди',
                    locale='uk',
                ),
                PlaceName(
                    name='Pays-Bas',
                    locale='fr',
                ),
            ],
            links={Link('https://en.wikipedia.org/wiki/Netherlands')},
        )
        self._load(netherlands)

        north_holland = Place(
            id='betty-demo-north-holland',
            names=[
                PlaceName(name='North Holland'),
                PlaceName(
                    name='Noord-Holland',
                    locale='nl',
                ),
                PlaceName(
                    name='Північна Голландія',
                    locale='uk',
                ),
                PlaceName(
                    name='Hollande-Septentrionale',
                    locale='fr',
                ),
            ],
            links={Link('https://en.wikipedia.org/wiki/North_Holland')},
        )
        self._load(Enclosure(encloses=north_holland, enclosed_by=netherlands))
        self._load(north_holland)

        amsterdam = Place(
            id='betty-demo-amsterdam',
            names=[
                PlaceName(name='Amsterdam'),
                PlaceName(
                    name='Амстерда́м',
                    locale='uk',
                ),
            ],
            coordinates=Point(52.366667, 4.9),
            links={Link('https://nl.wikipedia.org/wiki/Amsterdam')},
        )
        self._load(Enclosure(encloses=amsterdam, enclosed_by=north_holland))
        self._load(amsterdam)

        ilpendam = Place(
            id='betty-demo-ilpendam',
            names=[
                PlaceName(name='Ilpendam'),
                PlaceName(
                    name='Илпендам',
                    locale='uk',
                ),
            ],
            coordinates=Point(52.465556, 4.951111),
            links={Link('https://nl.wikipedia.org/wiki/Ilpendam')},
        )
        self._load(Enclosure(encloses=ilpendam, enclosed_by=north_holland))
        self._load(ilpendam)

        personal_accounts = Source(
            id='betty-demo-personal-accounts',
            name='Personal accounts',
        )
        self._load(personal_accounts)

        cite_first_person_account = Citation(
            id='betty-demo-first-person-account',
            source=personal_accounts,
        )
        self._load(cite_first_person_account)

        bevolkingsregister_amsterdam = Source(
            id='betty-demo-bevolkingsregister-amsterdam',
            name='Bevolkingsregister Amsterdam',
            author='Gemeente Amsterdam',
            publisher='Gemeente Amsterdam',
        )
        self._load(bevolkingsregister_amsterdam)

        david_marinus_lankester = Person(id='betty-demo-david-marinus-lankester')
        self._load(
            PersonName(
                person=david_marinus_lankester,
                individual='David Marinus',
                affiliation='Lankester',
            ),
            david_marinus_lankester,
        )

        geertruida_van_ling = Person(id='betty-demo-geertruida-van-ling')
        self._load(
            PersonName(
                person=geertruida_van_ling,
                individual='Geertruida',
                affiliation='Van Ling',
            ),
            geertruida_van_ling,
        )

        marriage_of_dirk_jacobus_lankester_and_jannigje_palsen = Event(
            id='betty-demo-marriage-of-dirk-jacobus-lankester-and-jannigje-palsen',
            event_type=Marriage,
            date=Date(1922, 7, 4),
            place=ilpendam,
        )
        self._load(marriage_of_dirk_jacobus_lankester_and_jannigje_palsen)

        birth_of_dirk_jacobus_lankester = Event(
            id='betty-demo-birth-of-dirk-jacobus-lankester',
            event_type=Birth,
            date=Date(1897, 8, 25),
            place=amsterdam,
        )
        self._load(birth_of_dirk_jacobus_lankester)

        death_of_dirk_jacobus_lankester = Event(
            id='betty-demo-death-of-dirk-jacobus-lankester',
            event_type=Death,
            date=Date(1986, 8, 18),
            place=amsterdam,
        )
        self._load(death_of_dirk_jacobus_lankester)

        dirk_jacobus_lankester = Person(
            id='betty-demo-dirk-jacobus-lankester',
            parents=(david_marinus_lankester, geertruida_van_ling)
        )
        self._load(
            PersonName(
                person=dirk_jacobus_lankester,
                individual='Dirk Jacobus',
                affiliation='Lankester',
            ),
            Presence(dirk_jacobus_lankester, Subject(), birth_of_dirk_jacobus_lankester),
            Presence(dirk_jacobus_lankester, Subject(), death_of_dirk_jacobus_lankester),
            Presence(dirk_jacobus_lankester, Subject(), marriage_of_dirk_jacobus_lankester_and_jannigje_palsen),
        )
        self._load(dirk_jacobus_lankester)

        birth_of_marinus_david_lankester = Event(
            id='betty-demo-birth-of-marinus-david',
            event_type=Birth,
            date=DateRange(Date(1874, 1, 15), Date(1874, 3, 21), start_is_boundary=True, end_is_boundary=True),
            place=amsterdam,
        )
        self._load(birth_of_marinus_david_lankester)

        death_of_marinus_david_lankester = Event(
            id='betty-demo-death-of-marinus-david',
            event_type=Death,
            date=Date(1971),
            place=amsterdam,
        )
        self._load(death_of_marinus_david_lankester)

        marinus_david_lankester = Person(
            id='betty-demo-marinus-david-lankester',
            parents=(david_marinus_lankester, geertruida_van_ling),
        )
        self._load(
            PersonName(
                person=marinus_david_lankester,
                individual='Marinus David',
                affiliation='Lankester',
            ),
            Presence(marinus_david_lankester, Subject(), birth_of_marinus_david_lankester),
            Presence(marinus_david_lankester, Subject(), death_of_marinus_david_lankester),
        )
        self._load(marinus_david_lankester)

        birth_of_jacoba_gesina_lankester = Event(
            id='betty-demo-birth-of-jacoba-gesina',
            event_type=Birth,
            date=Date(1900, 3, 14),
            place=amsterdam,
        )
        self._load(birth_of_jacoba_gesina_lankester)

        jacoba_gesina_lankester = Person(
            id='betty-demo-jacoba-gesina-lankester',
            parents=(david_marinus_lankester, geertruida_van_ling),
        )
        self._load(
            PersonName(
                person=jacoba_gesina_lankester,
                individual='Jacoba Gesina',
                affiliation='Lankester',
            ),
            Presence(jacoba_gesina_lankester, Subject(), birth_of_jacoba_gesina_lankester),
        )
        self._load(jacoba_gesina_lankester)

        jannigje_palsen = Person(id='betty-demo-jannigje-palsen')
        self._load(
            PersonName(
                person=jannigje_palsen,
                individual='Jannigje',
                affiliation='Palsen',
            ),
            Presence(jannigje_palsen, Subject(), marriage_of_dirk_jacobus_lankester_and_jannigje_palsen),
            jannigje_palsen,
        )

        marriage_of_johan_de_boer_and_liberta_lankester = Event(
            id='betty-demo-marriage-of-johan-de-boer-and-liberta-lankester',
            event_type=Marriage,
            date=Date(1953, 6, 19),
            place=amsterdam,
        )
        self._load(marriage_of_johan_de_boer_and_liberta_lankester)

        cite_birth_of_liberta_lankester_from_bevolkingsregister_amsterdam = Citation(
            id='betty-demo-birth-of-liberta-lankester-from-bevolkingsregister-amsterdam',
            source=bevolkingsregister_amsterdam,
            location=Str.plain('Amsterdam'),
        )
        self._load(cite_birth_of_liberta_lankester_from_bevolkingsregister_amsterdam)

        birth_of_liberta_lankester = Event(
            id='betty-demo-birth-of-liberta-lankester',
            event_type=Birth,
            date=Date(1929, 12, 22),
            place=amsterdam,
            citations=[cite_birth_of_liberta_lankester_from_bevolkingsregister_amsterdam],
        )
        self._load(birth_of_liberta_lankester)

        death_of_liberta_lankester = Event(
            id='betty-demo-death-of-liberta-lankester',
            event_type=Death,
            date=Date(2015, 1, 17),
            place=amsterdam,
            citations=[cite_first_person_account],
        )
        self._load(death_of_liberta_lankester)

        liberta_lankester = Person(
            id='betty-demo-liberta-lankester',
            parents=(dirk_jacobus_lankester, jannigje_palsen),
        )
        self._load(
            PersonName(
                person=liberta_lankester,
                individual='Liberta',
                affiliation='Lankester',
            ),
            PersonName(
                person=liberta_lankester,
                individual='Betty',
            ),
            Presence(liberta_lankester, Subject(), birth_of_liberta_lankester),
            Presence(liberta_lankester, Subject(), death_of_liberta_lankester),
            Presence(liberta_lankester, Subject(), marriage_of_johan_de_boer_and_liberta_lankester),
        )
        self._load(liberta_lankester)

        birth_of_johan_de_boer = Event(
            id='betty-demo-birth-of-johan-de-boer',
            event_type=Birth,
            date=Date(1930, 6, 20),
            place=amsterdam,
        )
        self._load(birth_of_johan_de_boer)

        death_of_johan_de_boer = Event(
            id='betty-demo-death-of-johan-de-boer',
            event_type=Death,
            date=Date(1999, 3, 10),
            place=amsterdam,
            citations=[cite_first_person_account],
        )
        self._load(death_of_johan_de_boer)

        johan_de_boer = Person(id='betty-demo-johan-de-boer')
        self._load(
            PersonName(
                person=johan_de_boer,
                individual='Johan',
                affiliation='De Boer',
            ),
            PersonName(
                person=johan_de_boer,
                individual='Hans',
            ),
            Presence(johan_de_boer, Subject(), birth_of_johan_de_boer),
            Presence(johan_de_boer, Subject(), death_of_johan_de_boer),
            Presence(johan_de_boer, Subject(), marriage_of_johan_de_boer_and_liberta_lankester),
            johan_de_boer,
        )

        parent_of_bart_feenstra_child_of_liberta_lankester = Person(
            id='betty-demo-parent-of-bart-feenstra-child-of-liberta-lankester',
            parents=(johan_de_boer, liberta_lankester),
        )
        self._load(PersonName(
            person=parent_of_bart_feenstra_child_of_liberta_lankester,
            individual='Bart\'s parent',
        ))
        self._load(parent_of_bart_feenstra_child_of_liberta_lankester)

        bart_feenstra = Person(
            id='betty-demo-bart-feenstra',
            parents=(parent_of_bart_feenstra_child_of_liberta_lankester,),
        )
        self._load(PersonName(
            person=bart_feenstra,
            individual='Bart',
            affiliation='Feenstra',
        ))
        self._load(bart_feenstra)

        theme = self.app.extensions[CottonCandy]
        theme.configuration.featured_entities.append(EntityReference(Person, 'betty-demo-liberta-lankester'))
        theme.configuration.featured_entities.append(EntityReference(Place, 'betty-demo-amsterdam'))


class DemoServer(Server):
    def __init__(self):
        self._app = App()
        super().__init__(self._app.localizer)
        self._app.project.configuration.extensions.append(ExtensionConfiguration(_Demo))
        # Include all of the translations Betty ships with.
        self._app.project.configuration.locales.replace(
            LocaleConfiguration('en-US', 'en'),
            LocaleConfiguration('nl-NL', 'nl'),
            LocaleConfiguration('fr-FR', 'fr'),
            LocaleConfiguration('uk', 'uk'),
        )
        self._server: Server | None = None
        self._exit_stack = AsyncExitStack()

    @classmethod
    def label(cls) -> Str:
        return Str._('Demo')

    @property
    def public_url(self) -> str:
        if self._server is not None:
            return self._server.public_url
        raise NoPublicUrlBecauseServerNotStartedError()

    async def start(self) -> None:
        try:
            await super().start()
            await self._exit_stack.enter_async_context(self._app)
            await load.load(self._app)
            self._server = AppServer.get(self._app)
            await self._exit_stack.enter_async_context(self._server)
            self._app.project.configuration.base_url = self._server.public_url
            await generate.generate(self._app)
        finally:
            await self.stop()

    async def stop(self) -> None:
        await self._exit_stack.aclose()
        await super().stop()



from __future__ import annotations

from pathlib import Path
from typing import Iterable, Any, Self

from reactives.instance.property import reactive_property

from betty.config import Configuration, ConfigurationSequence
from betty.serde.dump import minimize, Dump, VoidableDump
from betty.serde.load import Asserter, Fields, RequiredField, Assertions, OptionalField


class FamilyTreeConfiguration(Configuration):
    def __init__(self, file_path: Path | None = None):
        super().__init__()
        self.file_path = file_path

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FamilyTreeConfiguration):
            return False
        return self._file_path == other.file_path

    @property
    @reactive_property
    def file_path(self) -> Path | None:
        return self._file_path

    @file_path.setter
    def file_path(self, file_path: Path | None) -> None:
        self._file_path = file_path

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            RequiredField(
                'file',
                Assertions(asserter.assert_path()) | asserter.assert_setattr(configuration, 'file_path'),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return {
            'file': str(self.file_path),
        }


class FamilyTreeConfigurationSequence(ConfigurationSequence[FamilyTreeConfiguration]):
    def update(self, other: Self) -> None:
        self._clear_without_trigger()
        self.append(*other)

    @classmethod
    def _create_default_item(cls, configuration_key: int) -> FamilyTreeConfiguration:
        return FamilyTreeConfiguration()

    @classmethod
    def _item_type(cls) -> type[FamilyTreeConfiguration]:
        return FamilyTreeConfiguration


class GrampsConfiguration(Configuration):
    def __init__(self, family_trees: Iterable[FamilyTreeConfiguration] | None = None):
        super().__init__()
        self._family_trees = FamilyTreeConfigurationSequence(family_trees)
        self._family_trees.react(self)

    @property
    def family_trees(self) -> FamilyTreeConfigurationSequence:
        return self._family_trees

    def update(self, other: Self) -> None:
        self._family_trees.update(other._family_trees)

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            OptionalField(
                'family_trees',
                Assertions(configuration._family_trees.assert_load(configuration.family_trees)),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return minimize({
            'family_trees': self.family_trees.dump(),
        }, True)



from __future__ import annotations

from typing import TYPE_CHECKING

from betty.app.extension import ConfigurableExtension, UserFacingExtension
from betty.gramps.loader import GrampsLoader
from betty.locale import Str

if TYPE_CHECKING:
    from betty.extension.gramps.gui import _GrampsGuiWidget

from betty.extension.gramps.config import GrampsConfiguration
from betty.gui import GuiBuilder
from betty.load import Loader


class _Gramps(ConfigurableExtension[GrampsConfiguration], UserFacingExtension, Loader, GuiBuilder):
    @classmethod
    def default_configuration(cls) -> GrampsConfiguration:
        return GrampsConfiguration()

    async def load(self) -> None:
        for family_tree in self.configuration.family_trees:
            file_path = family_tree.file_path
            if file_path:
                await GrampsLoader(
                    self._app.project.ancestry,
                    localizer=self._app.localizer,
                ).load_file(file_path)

    @classmethod
    def label(cls) -> Str:
        return Str.plain('Gramps')

    @classmethod
    def description(cls) -> Str:
        return Str._('Load <a href="https://gramps-project.org/">Gramps</a> family trees.')

    def gui_build(self) -> _GrampsGuiWidget:
        from betty.extension.gramps.gui import _GrampsGuiWidget

        return _GrampsGuiWidget(self._app)



from __future__ import annotations

import logging
import re
from pathlib import Path
from shutil import copy2
from typing import Any, Callable, Iterable, cast, Self

from PyQt6.QtWidgets import QWidget
from aiofiles.os import makedirs
from jinja2 import pass_context
from jinja2.runtime import Context
from reactives.instance import ReactiveInstance
from reactives.instance.property import reactive_property

from betty.app.extension import ConfigurableExtension, Extension, Theme
from betty.config import Configuration
from betty.extension.cotton_candy.search import Index
from betty.extension.npm import _Npm, NpmBuilder, npm
from betty.functools import walk
from betty.generate import Generator, GenerationContext
from betty.gui import GuiBuilder
from betty.jinja2 import Jinja2Provider, context_app, context_localizer, context_task_context
from betty.locale import Date, Datey, Str
from betty.model import Entity, UserFacingEntity
from betty.model.ancestry import Event, Person, Presence, is_public
from betty.project import EntityReferenceSequence
from betty.serde.dump import minimize, Dump, VoidableDump
from betty.serde.load import AssertionFailed, Fields, Assertions, OptionalField, Asserter


class _ColorConfiguration(Configuration):
    _HEX_PATTERN = re.compile(r'^#[a-zA-Z0-9]{6}$')

    def __init__(self, hex_value: str):
        super().__init__()
        self._hex: str
        self.hex = hex_value

    def _validate_hex(self, hex_value: str) -> str:
        if not self._HEX_PATTERN.match(hex_value):
            raise AssertionFailed(Str._(
                '"{hex_value}" is not a valid hexadecimal color, such as #ffc0cb.',
                hex_value=hex_value,
            ))
        return hex_value

    @property
    @reactive_property
    def hex(self) -> str:
        return self._hex

    @hex.setter
    def hex(self, hex_value: str) -> None:
        if not self._HEX_PATTERN.match(hex_value):
            raise AssertionFailed(Str._(
                '"{hex_value}" is not a valid hexadecimal color, such as #ffc0cb.',
                hex_value=hex_value,
            ))
        self._hex = hex_value

    def update(self, other: Self) -> None:
        self.hex = other.hex

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        asserter = Asserter()
        hex_value = asserter.assert_str()(dump)
        if configuration is None:
            configuration = cls(hex_value)
        else:
            configuration.hex = hex_value
        return configuration

    def dump(self) -> VoidableDump:
        return self._hex


class CottonCandyConfiguration(Configuration):
    DEFAULT_PRIMARY_INACTIVE_COLOR = '#ffc0cb'
    DEFAULT_PRIMARY_ACTIVE_COLOR = '#ff69b4'
    DEFAULT_LINK_INACTIVE_COLOR = '#149988'
    DEFAULT_LINK_ACTIVE_COLOR = '#2a615a'

    def __init__(self):
        super().__init__()
        self._featured_entities = EntityReferenceSequence['UserFacingEntity & Entity']()
        self._featured_entities.react(self)
        self._primary_inactive_color = _ColorConfiguration(self.DEFAULT_PRIMARY_INACTIVE_COLOR)
        self._primary_inactive_color.react(self)
        self._primary_active_color = _ColorConfiguration(self.DEFAULT_PRIMARY_ACTIVE_COLOR)
        self._primary_active_color.react(self)
        self._link_inactive_color = _ColorConfiguration(self.DEFAULT_LINK_INACTIVE_COLOR)
        self._link_inactive_color.react(self)
        self._link_active_color = _ColorConfiguration(self.DEFAULT_LINK_ACTIVE_COLOR)
        self._link_active_color.react(self)

    @property
    def featured_entities(self) -> EntityReferenceSequence[UserFacingEntity & Entity]:
        return self._featured_entities

    @property
    def primary_inactive_color(self) -> _ColorConfiguration:
        return self._primary_inactive_color

    @property
    def primary_active_color(self) -> _ColorConfiguration:
        return self._primary_active_color

    @property
    def link_inactive_color(self) -> _ColorConfiguration:
        return self._link_inactive_color

    @property
    def link_active_color(self) -> _ColorConfiguration:
        return self._link_active_color

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            OptionalField(
                'featured_entities',
                Assertions(configuration._featured_entities.assert_load(configuration._featured_entities)),
            ),
            OptionalField(
                'primary_inactive_color',
                Assertions(configuration._primary_inactive_color.assert_load(configuration._primary_inactive_color)),
            ),
            OptionalField(
                'primary_active_color',
                Assertions(configuration._primary_active_color.assert_load(configuration._primary_active_color)),
            ),
            OptionalField(
                'link_inactive_color',
                Assertions(configuration._link_inactive_color.assert_load(configuration._link_inactive_color)),
            ),
            OptionalField(
                'link_active_color',
                Assertions(configuration._link_active_color.assert_load(configuration._link_active_color)),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return minimize({
            'featured_entities': self.featured_entities.dump(),
            'primary_inactive_color': self._primary_inactive_color.dump(),
            'primary_active_color': self._primary_active_color.dump(),
            'link_inactive_color': self._link_inactive_color.dump(),
            'link_active_color': self._link_active_color.dump(),
        })


class _CottonCandy(Theme, ConfigurableExtension[CottonCandyConfiguration], Generator, GuiBuilder, ReactiveInstance, NpmBuilder, Jinja2Provider):
    @classmethod
    def depends_on(cls) -> set[type[Extension]]:
        return {_Npm}

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / 'assets'

    @classmethod
    def label(cls) -> Str:
        return Str.plain('Cotton Candy')

    @classmethod
    def default_configuration(cls) -> CottonCandyConfiguration:
        return CottonCandyConfiguration()

    @classmethod
    def description(cls) -> Str:
        return Str._("Cotton Candy is Betty's default theme.")

    def gui_build(self) -> QWidget:
        from betty.extension.cotton_candy.gui import _CottonCandyGuiWidget

        return _CottonCandyGuiWidget(self._app)

    @property
    def globals(self) -> dict[str, Any]:
        return {
            'search_index': _global_search_index,
        }

    @property
    def filters(self) -> dict[str, Callable[..., Any]]:
        return {
            'person_timeline_events': lambda person: person_timeline_events(
                person,
                self.app.project.configuration.lifetime_threshold
            ),
        }

    async def npm_build(self, working_directory_path: Path, assets_directory_path: Path) -> None:
        await self.app.extensions[_Npm].install(type(self), working_directory_path)
        await npm(('run', 'webpack'), cwd=working_directory_path)
        await self._copy_npm_build(working_directory_path / 'webpack-build', assets_directory_path)
        logging.getLogger().info(self._app.localizer._('Built the Cotton Candy front-end assets.'))

    async def _copy_npm_build(self, source_directory_path: Path, destination_directory_path: Path) -> None:
        await makedirs(destination_directory_path, exist_ok=True)
        copy2(source_directory_path / 'cotton_candy.css', destination_directory_path / 'cotton_candy.css')
        copy2(source_directory_path / 'cotton_candy.js', destination_directory_path / 'cotton_candy.js')

    async def generate(self, task_context: GenerationContext) -> None:
        assets_directory_path = await self.app.extensions[_Npm].ensure_assets(self)
        await self._copy_npm_build(assets_directory_path, self.app.project.configuration.www_directory_path)


@pass_context
async def _global_search_index(context: Context) -> Iterable[dict[str, str]]:
    return await Index(
        context_app(context),
        context_task_context(context),
        context_localizer(context),
    ).build()


def _is_person_timeline_presence(presence: Presence) -> bool:
    if presence.private:
        return False
    if not presence.event:
        return False
    if not presence.event.date:
        return False
    if not presence.event.date.comparable:
        return False
    return True


def person_timeline_events(person: Person, lifetime_threshold: int) -> Iterable[Event]:
    # Collect all associated events for a person.
    # Start with the person's own events for which their presence is public.
    for presence in person.presences:
        if _is_person_timeline_presence(presence):
            yield cast(Event, presence.event)
        continue

    # If the person has start- or end-of-life events, we use those to constrain associated people's events.
    start_date = None
    end_date = None
    if person.start and _is_person_timeline_presence(person.start):
        start_date = cast(Event, person.start.event).date
    if person.end and _is_person_timeline_presence(person.end):
        end_date = cast(Event, person.end.event).date

    # If an end-of-life event exists, but no start-of-life event, create a start-of-life date based on the end date,
    # minus the lifetime threshold.
    if not start_date and end_date:
        if isinstance(end_date, Date):
            start_date_reference = end_date
        else:
            if end_date.end and end_date.end.comparable:
                start_date_reference = end_date.end
            else:
                start_date_reference = cast(Date, end_date.start)
        start_date = Date(
            cast(int, start_date_reference.year) - lifetime_threshold,
            start_date_reference.month,
            start_date_reference.day,
            start_date_reference.fuzzy,
        )

    # If a start-of-life event exists, but no end-of-life event, create an end-of-life date based on the start date,
    # plus the lifetime threshold.
    if not end_date and start_date:
        if isinstance(start_date, Date):
            end_date_reference = start_date
        else:
            if start_date.start and start_date.start.comparable:
                end_date_reference = start_date.start
            else:
                end_date_reference = cast(Date, start_date.end)
        end_date = Date(
            cast(int, end_date_reference.year) + lifetime_threshold,
            end_date_reference.month,
            end_date_reference.day,
            end_date_reference.fuzzy,
        )

    if not start_date or not end_date:
        reference_dates = list(sorted(
            cast(Datey, cast(Event, presence.event).date)
            for presence
            in person.presences
            if _is_person_timeline_presence(presence)
        ))
        if reference_dates:
            if not start_date:
                start_date = reference_dates[0]
            if not end_date:
                end_date = reference_dates[-1]

    if start_date and end_date:
        associated_people = filter(is_public, (
            # All ancestors.
            *walk(person, 'parents'),
            # All descendants.
            *walk(person, 'children'),
            # All siblings.
            *person.siblings,
        ))
        for associated_person in associated_people:
            # For associated events, we are only interested in people's start- or end-of-life events.
            for associated_presence in (associated_person.start, associated_person.end):
                if associated_presence is None:
                    continue
                if not _is_person_timeline_presence(associated_presence):
                    continue
                if cast(Event, associated_presence.event).date < start_date:
                    continue
                if cast(Event, associated_presence.event).date > end_date:
                    continue
                yield cast(Event, associated_presence.event)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, Callable, Any, Generic, TYPE_CHECKING, TypeVar, MutableSequence, MutableMapping, overload, \
    cast, TypeAlias

from betty.functools import _Result
from betty.locale import LocaleNotFoundError, get_data, Str
from betty.model import Entity, get_entity_type, EntityTypeImportError, EntityTypeInvalidError, EntityTypeError
from betty.serde.dump import DumpType, DumpTypeT, Void
from betty.serde.error import SerdeError, SerdeErrorCollection

if TYPE_CHECKING:
    from betty.app.extension import Extension

T = TypeVar('T')
ValueT = TypeVar('ValueT')
ReturnT = TypeVar('ReturnT')
ReturnU = TypeVar('ReturnU')
CallValueT = TypeVar('CallValueT')
CallReturnT = TypeVar('CallReturnT')
FValueT = TypeVar('FValueT')
MapReturnT = TypeVar('MapReturnT')
Number: TypeAlias = int | float
NumberT = TypeVar('NumberT', bound=Number)


class LoadError(SerdeError):
    pass


class AssertionFailed(LoadError):
    pass


class FormatError(LoadError):
    pass


Assertion: TypeAlias = Callable[
    [
        ValueT,
    ],
    ReturnT,
]


class _Assertions(Generic[CallValueT, CallReturnT, FValueT]):
    def __init__(self, _assertion: Assertion[FValueT, CallReturnT]):
        self._assertion = _assertion

    def extend(self, _assertion: Assertion[CallReturnT, MapReturnT]) -> _Assertions[CallValueT, MapReturnT, CallReturnT]:
        return _AssertionExtension(_assertion, self)

    def __or__(self, _assertion: Assertion[CallReturnT, MapReturnT]) -> _Assertions[CallValueT, MapReturnT, CallReturnT]:
        return self.extend(_assertion)

    def __call__(self, value: CallValueT) -> _Result[CallReturnT]:
        raise NotImplementedError(repr(self))


class Assertions(_Assertions[CallValueT, CallReturnT, CallValueT], Generic[CallValueT, CallReturnT]):
    def __call__(self, value: CallValueT) -> _Result[CallReturnT]:
        return _Result(value).map(self._assertion)


class _AssertionExtension(_Assertions[CallValueT, CallReturnT, FValueT], Generic[CallValueT, CallReturnT, FValueT]):
    def __init__(
        self,
        _assertion: Assertion[FValueT, CallReturnT],
        extended_assertion: _Assertions[CallValueT, FValueT, Any],
    ):
        super().__init__(_assertion)
        self._extended_assertion = extended_assertion

    def __call__(self, value: CallValueT) -> _Result[CallReturnT]:
        return self._extended_assertion(value).map(self._assertion)


@dataclass(frozen=True)
class _Field(Generic[ValueT, ReturnT]):
    name: str
    assertion: _Assertions[ValueT, ReturnT, Any] | None = None


@dataclass(frozen=True)
class RequiredField(Generic[ValueT, ReturnT], _Field[ValueT, ReturnT]):
    pass


@dataclass(frozen=True)
class OptionalField(Generic[ValueT, ReturnT], _Field[ValueT, ReturnT]):
    pass


class Fields:
    def __init__(self, *fields: _Field[Any, Any]):
        self._fields = fields

    def __iter__(self) -> Iterator[_Field[Any, Any]]:
        return (field for field in self._fields)


_AssertionBuilderFunction = Callable[[ValueT], ReturnT]
_AssertionBuilderMethod = Callable[[object, ValueT], ReturnT]
_AssertionBuilder = '_AssertionBuilderFunction[ValueT, ReturnT] | _AssertionBuilderMethod[ValueT, ReturnT]'


class Asserter:
    def _assert_type_violation_error_message(
            self,
            asserted_type: type[DumpType],
    ) -> Str:
        messages = {
            None: Str._('This must be none/null.'),
            bool: Str._('This must be a boolean.'),
            int: Str._('This must be a whole number.'),
            float: Str._('This must be a decimal number.'),
            str: Str._('This must be a string.'),
            list: Str._('This must be a list.'),
            dict: Str._('This must be a key-value mapping.'),
        }
        return messages[asserted_type]  # type: ignore[index]

    def _assert_type(
        self,
        value: Any,
        value_required_type: type[DumpTypeT],
        value_disallowed_type: type[DumpType] | None = None,
    ) -> DumpTypeT:
        if isinstance(value, value_required_type) and (value_disallowed_type is None or not isinstance(value, value_disallowed_type)):
            return value
        raise AssertionFailed(
            self._assert_type_violation_error_message(
                value_required_type,  # type: ignore[arg-type]
            )
        )

    def assert_or(self, if_assertion: Assertion[ValueT, ReturnT], else_assertions: Assertion[ValueT, ReturnU]) -> Assertion[ValueT, ReturnT | ReturnU]:
        def _assert_or(value: Any) -> ReturnT | ReturnU:
            assertions = (if_assertion, else_assertions)
            errors = SerdeErrorCollection()
            for assertion in assertions:
                try:
                    return assertion(value)
                except SerdeError as e:
                    if e.raised(AssertionFailed):
                        errors.append(e)
            raise errors
        return _assert_or

    def assert_none(self) -> Assertion[Any, None]:
        def _assert_none(value: Any) -> None:
            self._assert_type(value, type(None))
        return _assert_none

    def assert_bool(self) -> Assertion[Any, bool]:
        def _assert_bool(value: Any) -> bool:
            return self._assert_type(value, bool)
        return _assert_bool

    def assert_int(self) -> Assertion[Any, int]:
        def _assert_int(value: Any) -> int:
            return self._assert_type(value, int, bool)
        return _assert_int

    def assert_float(self) -> Assertion[Any, float]:
        def _assert_float(value: Any) -> float:
            return self._assert_type(value, float)
        return _assert_float

    def assert_number(self) -> Assertion[Any, int | float]:
        return self.assert_or(self.assert_int(), self.assert_float())

    def assert_positive_number(self) -> Assertion[Any, Number]:
        def _assert_positive_number(
            value: Any,
        ) -> Number:
            value = self.assert_number()(value)
            if value <= 0:
                raise AssertionFailed(Str._('This must be a positive number.'))
            return value
        return _assert_positive_number

    def assert_str(self) -> Assertion[Any, str]:
        def _assert_str(value: Any) -> str:
            return self._assert_type(value, str)
        return _assert_str

    def assert_list(self) -> Assertion[Any, list[Any]]:
        def _assert_list(value: Any) -> list[Any]:
            return self._assert_type(value, list)
        return _assert_list

    def assert_dict(self) -> Assertion[Any, dict[str, Any]]:
        def _assert_dict(value: Any) -> dict[str, Any]:
            return self._assert_type(value, dict)
        return _assert_dict

    def assert_assertions(self, assertions: _Assertions[ValueT, ReturnT, Any]) -> Assertion[Any, ReturnT]:
        def _assert_assertions(value: Any) -> ReturnT:
            return assertions(value).value
        return _assert_assertions

    def assert_sequence(self, item_assertion: Assertions[ValueT, ReturnT]) -> Assertion[Any, MutableSequence[ReturnT]]:
        def _assert_sequence(value: ValueT) -> MutableSequence[ReturnT]:
            list_value = self.assert_list()(value)
            sequence = []
            with SerdeErrorCollection().assert_valid() as errors:
                for value_item_index, value_item_value in enumerate(list_value):
                    with errors.catch(Str.plain(value_item_index)):
                        sequence.append(self.assert_assertions(item_assertion)(value_item_value))
            return sequence
        return _assert_sequence

    def assert_mapping(self, item_assertion: Assertions[ValueT, ReturnT]) -> Assertion[Any, MutableMapping[str, ReturnT]]:
        def _assert_mapping(value: ValueT) -> MutableMapping[str, ReturnT]:
            dict_value = self.assert_dict()(value)
            mapping = {}
            with SerdeErrorCollection().assert_valid() as errors:
                for value_item_key, value_item_value in dict_value.items():
                    with errors.catch(Str.plain(value_item_key)):
                        mapping[value_item_key] = self.assert_assertions(item_assertion)(value_item_value)
            return mapping
        return _assert_mapping

    def assert_fields(self, fields: Fields) -> Assertion[Any, MutableMapping[str, Any]]:
        def _assert_fields(value: Any) -> MutableMapping[str, Any]:
            value_dict = self.assert_dict()(value)
            mapping = {}
            with SerdeErrorCollection().assert_valid() as errors:
                for field in fields:
                    with errors.catch(Str.plain(field.name)):
                        if field.name in value_dict:
                            if field.assertion:
                                mapping[field.name] = self.assert_assertions(field.assertion)(value_dict[field.name])
                        elif isinstance(field, RequiredField):
                            raise AssertionFailed(Str._('This field is required.'))
            return mapping
        return _assert_fields

    @overload
    def assert_field(self, field: RequiredField[ValueT, ReturnT]) -> Assertion[ValueT, ReturnT]:
        pass

    @overload
    def assert_field(self, field: OptionalField[ValueT, ReturnT]) -> Assertion[ValueT, ReturnT | type[Void]]:
        pass

    def assert_field(self, field: _Field[ValueT, ReturnT]) -> Assertion[ValueT, ReturnT | type[Void]]:
        def _assert_field(value: Any) -> ReturnT | type[Void]:
            fields = self.assert_fields(Fields(field))(value)
            try:
                return cast('ReturnT | type[Void]', fields[field.name])
            except KeyError:
                if isinstance(field, RequiredField):
                    raise
                return Void
        return _assert_field

    def assert_record(self, fields: Fields) -> Assertion[Any, MutableMapping[str, Any]]:
        if not len(list(fields)):
            raise ValueError('One or more fields are required.')

        def _assert_record(value: Any) -> MutableMapping[str, Any]:
            dict_value = self.assert_dict()(value)
            known_keys = set(map(lambda x: x.name, fields))
            unknown_keys = set(dict_value.keys()) - known_keys
            with SerdeErrorCollection().assert_valid() as errors:
                for unknown_key in unknown_keys:
                    with errors.catch(Str.plain(unknown_key)):
                        raise AssertionFailed(Str._(
                            'Unknown key: {unknown_key}. Did you mean {known_keys}?',
                            unknown_key=f'"{unknown_key}"',
                            known_keys=', '.join(map(lambda x: f'"{x}"', sorted(known_keys)))
                        ))
                return self.assert_fields(fields)(dict_value)
        return _assert_record

    def assert_path(self) -> Assertion[Any, Path]:
        def _assert_path(value: Any) -> Path:
            self.assert_str()(value)
            return Path(value).expanduser().resolve()
        return _assert_path

    def assert_directory_path(self) -> Assertion[Any, Path]:
        def _assert_directory_path(value: Any) -> Path:
            directory_path = self.assert_path()(value)
            if directory_path.is_dir():
                return directory_path
            raise AssertionFailed(Str._(
                '"{path}" is not a directory.',
                path=value,
            ))
        return _assert_directory_path

    def assert_locale(self) -> Assertion[Any, str]:
        def _assert_locale(
            value: Any,
        ) -> str:
            value = self.assert_str()(value)
            try:
                get_data(value)
                return value
            except LocaleNotFoundError:
                raise AssertionFailed(Str._(
                    '"{locale}" is not a valid IETF BCP 47 language tag.',
                    locale=value,
                ))
        return _assert_locale

    def assert_setattr(self, instance: object, attr_name: str) -> Assertion[ValueT, ValueT]:
        def _assert_setattr(value: ValueT) -> ValueT:
            setattr(instance, attr_name, value)
            return value
        return _assert_setattr

    def assert_extension_type(self) -> Assertion[Any, type[Extension]]:
        def _assert_extension_type(
            value: Any,
        ) -> type[Extension]:
            from betty.app.extension import get_extension_type, ExtensionTypeImportError, ExtensionTypeInvalidError, ExtensionTypeError

            self.assert_str()(value)
            try:
                return get_extension_type(value)
            except ExtensionTypeImportError:
                raise AssertionFailed(
                    Str._(
                        'Cannot find and import "{extension_type}".',
                        extension_type=str(value),
                    )
                )
            except ExtensionTypeInvalidError:
                raise AssertionFailed(
                    Str._(
                        '"{extension_type}" is not a valid Betty extension type.',
                        extension_type=str(value),
                    )
                )
            except ExtensionTypeError:
                raise AssertionFailed(
                    Str._(
                        'Cannot determine the extension type for "{extension_type}". Did you perhaps make a typo, or could it be that the extension type comes from another package that is not yet installed?',
                        extension_type=str(value),
                    )
                )
        return _assert_extension_type

    def assert_entity_type(self) -> Assertion[Any, type[Entity]]:
        def _assert_entity_type(
            value: Any,
        ) -> type[Entity]:
            self.assert_str()(value)
            try:
                return get_entity_type(value)
            except EntityTypeImportError:
                raise AssertionFailed(
                    Str._(
                        'Cannot find and import "{entity_type}".',
                        entity_type=str(value),
                    )
                )
            except EntityTypeInvalidError:
                raise AssertionFailed(
                    Str._(
                        '"{entity_type}" is not a valid Betty entity type.',
                        entity_type=str(value),
                    )
                )
            except EntityTypeError:
                raise AssertionFailed(
                    Str._(
                        'Cannot determine the entity type for "{entity_type}". Did you perhaps make a typo, or could it be that the entity type comes from another package that is not yet installed?',
                        entity_type=str(value),
                    )
                )
        return _assert_entity_type



from __future__ import annotations

import asyncio
import logging
import sys
from contextlib import suppress, contextmanager
from functools import wraps
from pathlib import Path
from typing import Callable, TypeVar, cast, Iterator, Awaitable, ParamSpec, Concatenate

import click
from PyQt6.QtWidgets import QMainWindow
from click import get_current_context, Context, Option, Command, Parameter

from betty import about, generate, load
from betty.app import App
from betty.asyncio import sync, wait
from betty.error import UserFacingError
from betty.extension import demo
from betty.gui import BettyApplication
from betty.gui.app import WelcomeWindow
from betty.gui.project import ProjectWindow
from betty.locale import update_translations, init_translation, Str
from betty.logging import CliHandler
from betty.serde.load import AssertionFailed
from betty.serve import AppServer

T = TypeVar('T')
P = ParamSpec('P')


class CommandProvider:
    @property
    def commands(self) -> dict[str, Command]:
        raise NotImplementedError(repr(self))


@contextmanager
def catch_exceptions() -> Iterator[None]:
    try:
        yield
    except KeyboardInterrupt:
        print('Quitting...')
        sys.exit(0)
        pass
    except Exception as e:
        logger = logging.getLogger()
        if isinstance(e, UserFacingError):
            logger.error(str(e))
        else:
            logger.exception(e)
        sys.exit(1)


def _command(
    f: Callable[P, Awaitable[None]] | Callable[Concatenate[App, P], Awaitable[None]],
    is_app_command: bool,
) -> Callable[P, None]:
    @wraps(f)
    @catch_exceptions()
    def _command(*args: P.args, **kwargs: P.kwargs) -> None:
        if is_app_command:
            # We must get the current Click context from the main thread.
            # Once that is done, we can wait for the async commands to complete, which MAY be done in a thread.
            app = get_current_context().obj['app']

            async def _app_command():
                async with app:
                    await f(app, *args, **kwargs)
            return wait(_app_command())
        return wait(f(*args, **kwargs))
    return _command


def global_command(f: Callable[P, Awaitable[None]]) -> Callable[P, None]:
    return _command(f, False)


def app_command(f: Callable[Concatenate[App, P], Awaitable[None]]) -> Callable[P, None]:
    return _command(f, True)


@catch_exceptions()
@sync
async def _init_ctx(
    ctx: Context,
    __: Option | Parameter | None = None,
    configuration_file_path: str | None = None,
) -> None:
    ctx.ensure_object(dict)

    if 'initialized' in ctx.obj:
        return
    ctx.obj['initialized'] = True

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.addHandler(CliHandler())

    app = App()
    ctx.obj['commands'] = {
        'clear-caches': _clear_caches,
        'demo': _demo,
        'gui': _gui,
    }
    if about.is_development():
        ctx.obj['commands']['init-translation'] = _init_translation
        ctx.obj['commands']['update-translations'] = _update_translations
    ctx.obj['app'] = app

    if configuration_file_path is None:
        try_configuration_file_paths = [
            Path.cwd() / f'betty{extension}'
            for extension
            in {'.json', '.yaml', '.yml'}
        ]
    else:
        try_configuration_file_paths = [Path.cwd() / configuration_file_path]

    async with app:
        for try_configuration_file_path in try_configuration_file_paths:
            try:
                await app.project.configuration.read(try_configuration_file_path)
            except FileNotFoundError:
                continue
            else:
                ctx.obj['commands']['generate'] = _generate
                ctx.obj['commands']['serve'] = _serve
                for extension in app.extensions.flatten():
                    if isinstance(extension, CommandProvider):
                        for command_name, command in extension.commands.items():
                            ctx.obj['commands'][command_name] = command
                logger.info(app.localizer._('Loaded the configuration from {configuration_file_path}.').format(
                    configuration_file_path=str(try_configuration_file_path)),
                )
                return

        if configuration_file_path is not None:
            raise AssertionFailed(Str._(
                'Configuration file "{configuration_file_path}" does not exist.',
                configuration_file_path=configuration_file_path,
            ))


class _BettyCommands(click.MultiCommand):
    @catch_exceptions()
    def list_commands(self, ctx: Context) -> list[str]:
        _init_ctx(ctx)
        return list(ctx.obj['commands'].keys())

    @catch_exceptions()
    def get_command(self, ctx: Context, cmd_name: str) -> Command | None:
        _init_ctx(ctx)
        with suppress(KeyError):
            return cast(Command, ctx.obj['commands'][cmd_name])
        return None


@click.command(cls=_BettyCommands)
@click.option(
    '--configuration',
    '-c',
    'app',
    is_eager=True,
    help='The path to a Betty project configuration file. Defaults to betty.json|yaml|yml in the current working directory. This will make additional commands available.',
    callback=_init_ctx,
)
@click.version_option(about.version_label(), message=about.report(), prog_name='Betty')
def main(app: App) -> None:
    pass


@click.command(help='Clear all caches.')
@global_command
async def _clear_caches() -> None:
    async with App() as app:
        await app.cache.clear()


@click.command(help='Explore a demonstration site.')
@global_command
async def _demo() -> None:
    async with demo.DemoServer() as server:
        await server.show()
        while True:
            await asyncio.sleep(999)


@click.command(help="Open Betty's graphical user interface (GUI).")
@click.option(
    '--configuration',
    '-c',
    'configuration_file_path',
    is_eager=True,
    help='The path to a Betty project configuration file. Defaults to betty.json|yaml|yml in the current working directory.',
    callback=lambda _, __, configuration_file_path: Path(configuration_file_path) if configuration_file_path else None,
)
@global_command
async def _gui(configuration_file_path: Path | None) -> None:
    async with App() as app:
        qapp = BettyApplication([sys.argv[0]], app=app)
        window: QMainWindow
        if configuration_file_path is None:
            window = WelcomeWindow(app)
        else:
            await app.project.configuration.read(configuration_file_path)
            window = ProjectWindow(app)
        window.show()
        sys.exit(qapp.exec())


@click.command(help='Generate a static site.')
@app_command
async def _generate(app: App) -> None:
    await load.load(app)
    await generate.generate(app)


@click.command(help='Serve a generated site.')
@app_command
async def _serve(app: App) -> None:
    async with AppServer.get(app) as server:
        await server.show()
        while True:
            await asyncio.sleep(999)


if about.is_development():
    @click.command(short_help='Initialize a new translation', help='Initialize a new translation.\n\nThis is available only when developing Betty.')
    @click.argument('locale')
    @global_command
    async def _init_translation(locale: str) -> None:
        await init_translation(locale)

    @click.command(short_help='Update all existing translations', help='Update all existing translations.\n\nThis is available only when developing Betty.')
    @global_command
    async def _update_translations() -> None:
        await update_translations()



from __future__ import annotations

from pathlib import Path
from typing import Callable, Iterable, Any

from jinja2 import pass_context
from jinja2.runtime import Context
from reactives.instance import ReactiveInstance
from reactives.instance.property import reactive_property

from betty.app.extension import UserFacingExtension
from betty.asyncio import gather
from betty.jinja2 import Jinja2Provider, context_localizer
from betty.load import PostLoader
from betty.locale import negotiate_locale, Str
from betty.model.ancestry import Link
from betty.wikipedia import _Retriever, _Populator, Entry, _parse_url, NotAnEntryError, RetrievalError


class _Wikipedia(UserFacingExtension, Jinja2Provider, PostLoader, ReactiveInstance):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.__retriever: _Retriever | None = None
        self.__populator: _Populator | None = None

    async def post_load(self) -> None:
        await self._populator.populate()

    @property
    @reactive_property(on_trigger_delete=True)
    def _retriever(self) -> _Retriever:
        if self.__retriever is None:
            self.__retriever = _Retriever(self.app.http_client, self.cache_directory_path)
        return self.__retriever

    @_retriever.deleter
    def _retriever(self) -> None:
        self.__retriever = None

    @property
    @reactive_property(on_trigger_delete=True)
    def _populator(self) -> _Populator:
        if self.__populator is None:
            self.__populator = _Populator(self.app, self._retriever)
        return self.__populator

    @_populator.deleter
    def _populator(self) -> None:
        self.__populator = None

    @property
    def filters(self) -> dict[str, Callable[..., Any]]:
        return {
            'wikipedia': self._filter_wikipedia_links,
        }

    @pass_context
    async def _filter_wikipedia_links(self, context: Context, links: Iterable[Link]) -> Iterable[Entry]:
        return filter(
            None,
            await gather(*(
                self._filter_wikipedia_link(
                    context_localizer(context).locale,
                    link,
                )
                for link
                in links
            )),
        )

    async def _filter_wikipedia_link(self, locale: str, link: Link) -> Entry | None:
        try:
            entry_language, entry_name = _parse_url(link.url)
        except NotAnEntryError:
            return None
        if negotiate_locale(locale, {entry_language}) is None:
            return None
        try:
            return await self._retriever.get_entry(entry_language, entry_name)
        except RetrievalError:
            return None

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / 'assets'

    @classmethod
    def label(cls) -> Str:
        return Str._('Wikipedia')

    @classmethod
    def description(cls) -> Str:
        return Str._("""
Display <a href="https://www.wikipedia.org/">Wikipedia</a> summaries for resources with external links. In your custom <a href="https://jinja2docs.readthedocs.io/en/stable/">Jinja2</a> templates, use the following: <pre><code>
{% with resource=resource_with_links %}
    {% include 'wikipedia.html.j2' %}
{% endwith %}
</code></pre>""")



"""
This is a partial, simplified fork of the abandoned python-resize-image package.

This module exists purely because existing Betty code depended on an abandoned package, and that code has not been
refactored yet.

New code MUST NOT use this module.
"""
import math

from PIL.Image import Image


def resize_crop(image: Image, size: tuple[int, int]) -> Image:
    img_format = image.format
    image = image.copy()
    old_size = image.size
    left = (old_size[0] - size[0]) / 2
    top = (old_size[1] - size[1]) / 2
    right = old_size[0] - left
    bottom = old_size[1] - top
    rect = [int(math.ceil(x)) for x in (left, top, right, bottom)]
    left, top, right, bottom = rect
    crop = image.crop((left, top, right, bottom))
    crop.format = img_format
    return crop


def resize_cover(image: Image, size: tuple[int, int]) -> Image:
    img_format = image.format
    img = image.copy()
    img_size = img.size
    ratio = max(size[0] / img_size[0], size[1] / img_size[1])
    new_size = [
        int(math.ceil(img_size[0] * ratio)),
        int(math.ceil(img_size[1] * ratio))
    ]
    img = img.resize((new_size[0], new_size[1]))
    img = resize_crop(img, size)
    img.format = img_format
    return img


def resize_width(image: Image, width: int) -> Image:
    img_format = image.format
    img = image.copy()
    img_size = img.size
    # If the origial image has already the good width, return it
    # fix issue #16
    if img_size[0] == width:
        return image
    new_height = int(math.ceil((width / img_size[0]) * img_size[1]))
    img.thumbnail((width, new_height))
    img.format = img_format
    return img


def resize_height(image: Image, height: int) -> Image:
    img_format = image.format
    img = image.copy()
    img_size = img.size
    # If the origial image has already the good height, return it
    # fix issue #16
    if img_size[1] == height:
        return image
    new_width = int(math.ceil((height / img_size[1]) * img_size[0]))
    img.thumbnail((new_width, height))
    img.format = img_format
    return img



from __future__ import annotations

import logging
import subprocess
from pathlib import Path
from shutil import copy2

from aiofiles.os import makedirs

from betty.app.extension import Extension, UserFacingExtension
from betty.cache import CacheScope
from betty.extension.npm import _Npm, NpmBuilder, npm
from betty.generate import Generator, GenerationContext
from betty.html import CssProvider, JsProvider
from betty.locale import Str


class _Trees(UserFacingExtension, CssProvider, JsProvider, Generator, NpmBuilder):
    @classmethod
    def depends_on(cls) -> set[type[Extension]]:
        return {_Npm}

    async def npm_build(self, working_directory_path: Path, assets_directory_path: Path) -> None:
        await self.app.extensions[_Npm].install(type(self), working_directory_path)
        await npm(('run', 'webpack'), cwd=working_directory_path, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        await self._copy_npm_build(working_directory_path / 'webpack-build', assets_directory_path)
        logging.getLogger().info(self._app.localizer._('Built the interactive family trees.'))

    async def _copy_npm_build(self, source_directory_path: Path, destination_directory_path: Path) -> None:
        await makedirs(destination_directory_path, exist_ok=True)
        copy2(source_directory_path / 'trees.css', destination_directory_path / 'trees.css')
        copy2(source_directory_path / 'trees.js', destination_directory_path / 'trees.js')

    @classmethod
    def npm_cache_scope(cls) -> CacheScope:
        return CacheScope.BETTY

    async def generate(self, task_context: GenerationContext) -> None:
        assets_directory_path = await self.app.extensions[_Npm].ensure_assets(self)
        await self._copy_npm_build(assets_directory_path, self.app.project.configuration.www_directory_path)

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / 'assets'

    @property
    def public_css_paths(self) -> list[str]:
        return [
            self.app.static_url_generator.generate('trees.css'),
        ]

    @property
    def public_js_paths(self) -> list[str]:
        return [
            self.app.static_url_generator.generate('trees.js'),
        ]

    @classmethod
    def label(cls) -> Str:
        return Str._('Trees')

    @classmethod
    def description(cls) -> Str:
        return Str._('Display interactive family trees using <a href="https://cytoscape.org/">Cytoscape</a>.')



from __future__ import annotations

from typing import Any

from PyQt6.QtCore import Qt, QThread, pyqtSignal
from PyQt6.QtWidgets import QVBoxLayout, QWidget, QPushButton

from betty.app import App
from betty.asyncio import sync
from betty.gui import BettyWindow
from betty.gui.error import catch_exceptions
from betty.gui.text import Text
from betty.project import Project
from betty.serve import Server, AppServer


class _ServeThread(QThread):
    server_started = pyqtSignal()

    def __init__(self, project: Project, server_name: str, serve_window: _ServeWindow, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._project = project
        self._server_name = server_name
        self._serve_window = serve_window
        self._app: App | None = None
        self._server: Server | None = None

    @property
    def server(self) -> Server:
        if self._server is None:
            raise RuntimeError('This thread has not been started yet.')
        return self._server

    @sync
    async def run(self) -> None:
        with catch_exceptions(parent=self._serve_window, close_parent=True):
            async with App(project=self._project) as self._app:
                self._server = self._app.servers[self._server_name]
                await self._server.start()
                self.server_started.emit()
                await self._server.show()

    @sync
    async def stop(self) -> None:
        server = self._server
        if server is not None:
            await server.stop()


class _ServeWindow(BettyWindow):
    window_width = 500
    window_height = 100

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self.__thread: _ServeThread | None = None

        self._central_layout = QVBoxLayout()
        central_widget = QWidget()
        central_widget.setLayout(self._central_layout)
        self.setCentralWidget(central_widget)

        self._loading_instruction = Text(self._app.localizer._('Loading...'))
        self._loading_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._central_layout.addWidget(self._loading_instruction)

    def _server_name(self) -> str:
        raise NotImplementedError(repr(self))

    @property
    def _thread(self) -> _ServeThread:
        if self.__thread is None:
            raise RuntimeError('This window has not been shown yet.')
        return self.__thread

    def _build_instruction(self) -> str:
        raise NotImplementedError(repr(self))

    def _server_started(self) -> None:
        # The server may have been stopped before this method was called.
        if self.__thread is None:
            return

        self._loading_instruction.close()

        instance_instruction = Text(self._build_instruction())
        instance_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._central_layout.addWidget(instance_instruction)

        general_instruction = Text(self._app.localizer._('Keep this window open to keep the site running.'))
        general_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._central_layout.addWidget(general_instruction)

        stop_server_button = QPushButton(self._app.localizer._('Stop the site'), self)
        stop_server_button.released.connect(
            self.close,
        )
        self._central_layout.addWidget(stop_server_button)

    def show(self) -> None:
        super().show()
        # Explicitly activate this window in case it existed and was shown before, but requested again.
        self.activateWindow()
        self._start()

    def _start(self) -> None:
        if self.__thread is None:
            self.__thread = _ServeThread(self._app.project, self._server_name(), self)
            self.__thread.server_started.connect(self._server_started)
            self.__thread.start()

    def close(self) -> bool:
        self._stop()
        return super().close()

    def _stop(self) -> None:
        if self.__thread is not None:
            self.__thread.stop()
        self.__thread = None


class ServeProjectWindow(_ServeWindow):
    def _server_name(self) -> str:
        return AppServer.get(self._app).name()

    @property
    def title(self) -> str:
        return self._app.localizer._('Serving your site...')

    def _build_instruction(self) -> str:
        return self._app.localizer._('You can now view your site at <a href="{url}">{url}</a>.').format(
            url=self._thread.server.public_url,
        )


class ServeDemoWindow(_ServeWindow):
    def _server_name(self) -> str:
        from betty.extension import demo

        return demo.DemoServer.name()

    def _build_instruction(self) -> str:
        return self._app.localizer._('You can now view a Betty demonstration site at <a href="{url}">{url}</a>.').format(
            url=self._thread.server.public_url,
        )

    @property
    def title(self) -> str:
        return self._app.localizer._('Serving the Betty demo...')



from __future__ import annotations

from collections import defaultdict

from betty.app.extension import UserFacingExtension
from betty.load import PostLoader, getLogger
from betty.locale import Str
from betty.model import Entity
from betty.model.ancestry import Person, HasMutablePrivacy
from betty.privatizer import Privatizer as PrivatizerApi


class _Privatizer(UserFacingExtension, PostLoader):
    async def post_load(self) -> None:
        self.privatize()

    @classmethod
    def label(cls) -> Str:
        return Str._('Privatizer')

    @classmethod
    def description(cls) -> Str:
        return Str._('Determine if people can be proven to have died. If not, mark them and their associated entities private.')

    def privatize(self) -> None:
        logger = getLogger()
        logger.info(self._app.localizer._('Privatizing...'))

        privatizer = PrivatizerApi(self._app.project.configuration.lifetime_threshold)

        newly_privatized: dict[type[HasMutablePrivacy & Entity], int] = defaultdict(lambda: 0)
        entities: list[HasMutablePrivacy & Entity] = []
        for entity in self._app.project.ancestry:
            if isinstance(entity, HasMutablePrivacy):
                entities.append(entity)
                if entity.private:
                    newly_privatized[
                        entity.type  # type: ignore[index]
                    ] -= 1

        for entity in entities:
            privatizer.privatize(entity)

        for entity in entities:
            if entity.private:
                newly_privatized[entity.type] += 1  # type: ignore[index]

        if newly_privatized[Person] > 0:
            logger.info(self._app.localizer._('Privatized {count} people because they are likely still alive.').format(
                count=str(newly_privatized[Person]),
            ))
        for entity_type in set(newly_privatized) - {Person}:
            if newly_privatized[entity_type] > 0:
                logger.info(self._app.localizer._('Privatized {count} {entity_type}, because they are associated with private people.').format(
                    count=str(newly_privatized[entity_type]),
                    entity_type=entity_type.entity_type_label_plural().localize(self._app.localizer),
                ))



from __future__ import annotations

import functools
from collections import defaultdict
from importlib.metadata import entry_points, EntryPoint
from pathlib import Path
from typing import Any, TypeVar, Iterable, TYPE_CHECKING, Generic, \
    Iterator, Sequence, Self

from reactives import scope
from reactives.instance import ReactiveInstance

from betty import fs
from betty.app.extension.requirement import Requirement
from betty.asyncio import gather
from betty.config import ConfigurationT, Configurable
from betty.dispatch import Dispatcher, TargetedDispatcher
from betty.importlib import import_any
from betty.locale import Str

if TYPE_CHECKING:
    from betty.app import App


class ExtensionError(BaseException):
    pass


class ExtensionTypeError(ExtensionError, ValueError):
    pass


class ExtensionTypeImportError(ExtensionTypeError, ImportError):
    """
    Raised when an alleged extension type cannot be imported.
    """
    def __init__(self, extension_type_name: str):
        super().__init__(f'Cannot find and import an extension with name "{extension_type_name}".')


class ExtensionTypeInvalidError(ExtensionTypeError, ImportError):
    """
    Raised for types that are not valid extension types.
    """
    def __init__(self, extension_type: type):
        super().__init__(f'{extension_type.__module__}.{extension_type.__name__} is not an extension type class. Extension types must extend {Extension.__module__}.{Extension.__name__}.')


class CyclicDependencyError(ExtensionError, RuntimeError):
    def __init__(self, extension_types: Iterable[type[Extension]]):
        extension_names = ', '.join([extension.name() for extension in extension_types])
        super().__init__(f'The following extensions have cyclic dependencies: {extension_names}')


class Dependencies(Requirement):
    def __init__(self, dependent_type: type[Extension]):
        super().__init__()
        self._dependent_type = dependent_type

    @classmethod
    def for_dependent(cls, dependent_type: type[Extension]) -> Self:
        return cls(dependent_type)

    def is_met(self) -> bool:
        for dependency_type in self._dependent_type.depends_on():
            try:
                if not dependency_type.enable_requirement().is_met():
                    return False
            except RecursionError:
                raise CyclicDependencyError([dependency_type])
        return True

    def summary(self) -> Str:
        return Str._(
            '{dependent_label} requires {dependency_labels}.',
            dependent_label=format_extension_type(self._dependent_type),
            dependency_labels=', '.join(map(format_extension_type, self._dependent_type.depends_on())),
        )


class Dependents(Requirement):
    def __init__(self, dependency: Extension, dependents: Iterable[Extension]):
        super().__init__()
        self._dependency = dependency
        self._dependents = dependents

    def summary(self) -> Str:
        return Str._(
            '{dependency_label} is required by {dependency_labels}.',
            dependency_label=format_extension_type(type(self._dependency)),
            dependent_labels=', '.join(
                map(
                    format_extension_type,
                    map(type, self._dependents),
                )
            ),
        )

    def is_met(self) -> bool:
        # This class is never instantiated unless there is at least one enabled dependent, which means this requirement
        # is always met.
        return True

    @classmethod
    def for_dependency(cls, dependency: Extension) -> Self:
        dependents = [
            dependency.app.extensions[extension_type]
            for extension_type
            in discover_extension_types()
            if dependency.__class__ in extension_type.depends_on() and extension_type in dependency.app.extensions
        ]
        return cls(dependency, dependents)


class Extension:
    """
    Integrate optional functionality with the Betty app.
    """

    def __init__(self, app: App, *args: Any, **kwargs: Any):
        assert type(self) is not Extension
        super().__init__(*args, **kwargs)
        self._app = app

    @classmethod
    def name(cls) -> str:
        return '%s.%s' % (cls.__module__, cls.__name__)

    @classmethod
    def depends_on(cls) -> set[type[Extension]]:
        return set()

    @classmethod
    def comes_after(cls) -> set[type[Extension]]:
        return set()

    @classmethod
    def comes_before(cls) -> set[type[Extension]]:
        return set()

    @classmethod
    def enable_requirement(cls) -> Requirement:
        """
        Define the requirement for this extension to be enabled.

        This defaults to the extension's dependencies.
        """
        return Dependencies.for_dependent(cls)

    def disable_requirement(self) -> Requirement:
        """
        Define the requirement for this extension to be disabled.

        This defaults to the extension's dependents.
        """
        return Dependents.for_dependency(self)

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return None

    @property
    def cache_directory_path(self) -> Path:
        return fs.CACHE_DIRECTORY_PATH / self.name()

    @property
    def app(self) -> App:
        return self._app


ExtensionT = TypeVar('ExtensionT', bound=Extension)


class UserFacingExtension(Extension):
    @classmethod
    def label(cls) -> Str:
        raise NotImplementedError(repr(cls))

    @classmethod
    def description(cls) -> Str:
        raise NotImplementedError(repr(cls))


class Theme(UserFacingExtension):
    pass


@functools.singledispatch
def get_extension_type(extension_type_definition: str | type[Extension] | Extension) -> type[Extension]:
    raise ExtensionTypeError(f'Cannot get the extension type for "{extension_type_definition}".')


@get_extension_type.register(str)
def get_extension_type_by_name(extension_type_name: str) -> type[Extension]:
    try:
        extension_type = import_any(extension_type_name)
    except ImportError:
        raise ExtensionTypeImportError(extension_type_name) from None
    return get_extension_type(extension_type)


@get_extension_type.register(type)
def get_extension_type_by_type(extension_type: type) -> type[Extension]:
    if issubclass(extension_type, Extension):
        return extension_type
    raise ExtensionTypeInvalidError(extension_type)


@get_extension_type.register(Extension)
def get_extension_type_by_extension(extension: Extension) -> type[Extension]:
    return get_extension_type(type(extension))


def format_extension_type(extension_type: type[Extension]) -> str:
    if issubclass(extension_type, UserFacingExtension):
        return f'{extension_type.label()} ({extension_type.name()})'
    return extension_type.name()


class ConfigurableExtension(Extension, Generic[ConfigurationT], Configurable[ConfigurationT]):
    def __init__(self, *args: Any, configuration: ConfigurationT | None = None, **kwargs: Any):
        assert type(self) is not ConfigurableExtension
        super().__init__(*args, **kwargs)
        self._configuration = configuration or self.default_configuration()

    @classmethod
    def default_configuration(cls) -> ConfigurationT:
        raise NotImplementedError(repr(cls))


class Extensions(ReactiveInstance):
    def __getitem__(self, extension_type: type[ExtensionT] | str) -> ExtensionT:
        raise NotImplementedError(repr(self))

    def __iter__(self) -> Iterator[Iterator[Extension]]:
        raise NotImplementedError(repr(self))

    def flatten(self) -> Iterator[Extension]:
        raise NotImplementedError(repr(self))

    def __contains__(self, extension_type: type[Extension] | str | Any) -> bool:
        raise NotImplementedError(repr(self))


class ListExtensions(Extensions):
    def __init__(self, extensions: list[list[Extension]]):
        super().__init__()
        self._extensions = extensions

    @scope.register_self
    def __getitem__(self, extension_type: type[ExtensionT] | str) -> ExtensionT:
        if isinstance(extension_type, str):
            extension_type = import_any(extension_type)
        for extension in self.flatten():
            if type(extension) is extension_type:
                return extension  # type: ignore[return-value]
        raise KeyError(f'Unknown extension of type "{extension_type}"')

    @scope.register_self
    def __iter__(self) -> Iterator[Iterator[Extension]]:
        # Use a generator so we discourage calling code from storing the result.
        for batch in self._extensions:
            yield (extension for extension in batch)

    def flatten(self) -> Iterator[Extension]:
        for batch in self:
            yield from batch

    @scope.register_self
    def __contains__(self, extension_type: type[Extension] | str) -> bool:
        if isinstance(extension_type, str):
            try:
                extension_type = import_any(extension_type)
            except ImportError:
                return False
        for extension in self.flatten():
            if type(extension) is extension_type:
                return True
        return False


class ExtensionDispatcher(Dispatcher):
    def __init__(self, extensions: Extensions):
        self._extensions = extensions

    def dispatch(self, target_type: type[Any]) -> TargetedDispatcher:
        target_method_names = [method_name for method_name in dir(target_type) if not method_name.startswith('_')]
        if len(target_method_names) != 1:
            raise ValueError(f"A dispatch's target type must have a single method to dispatch to, but {target_type} has {len(target_method_names)}.")
        target_method_name = target_method_names[0]

        async def _dispatch(*args: Any, **kwargs: Any) -> list[Any]:
            return [
                result
                for target_extension_batch
                in self._extensions
                for result
                in await gather(*(
                    getattr(target_extension, target_method_name)(*args, **kwargs)
                    for target_extension in target_extension_batch
                    if isinstance(target_extension, target_type)
                ))
            ]
        return _dispatch


ExtensionTypeGraph = dict[type[Extension], set[type[Extension]]]


def build_extension_type_graph(extension_types: Iterable[type[Extension]]) -> ExtensionTypeGraph:
    extension_types_graph: ExtensionTypeGraph = defaultdict(set)
    # Add dependencies to the extension graph.
    for extension_type in extension_types:
        _extend_extension_type_graph(extension_types_graph, extension_type)
    # Now all dependencies have been collected, extend the graph with optional extension orders.
    for extension_type in extension_types:
        for before in extension_type.comes_before():
            if before in extension_types_graph:
                extension_types_graph[before].add(extension_type)
        for after in extension_type.comes_after():
            if after in extension_types_graph:
                extension_types_graph[extension_type].add(after)

    return extension_types_graph


def _extend_extension_type_graph(graph: ExtensionTypeGraph, extension_type: type[Extension]) -> None:
    dependencies = extension_type.depends_on()
    # Ensure each extension type appears in the graph, even if they're isolated.
    graph.setdefault(extension_type, set())
    for dependency in dependencies:
        seen_dependency = dependency in graph
        graph[extension_type].add(dependency)
        if not seen_dependency:
            _extend_extension_type_graph(graph, dependency)


def discover_extension_types() -> set[type[Extension]]:
    betty_entry_points: Sequence[EntryPoint]
    betty_entry_points = entry_points(  # type: ignore[assignment, unused-ignore]
        group='betty.extensions',  # type: ignore[call-arg, unused-ignore]
    )
    return {import_any(betty_entry_point.value) for betty_entry_point in betty_entry_points}



from __future__ import annotations

import json as stdjson
from os import path
from pathlib import Path
from typing import Any, Callable, TypeVar, cast

import jsonschema
from geopy import Point
from jsonschema import RefResolver

from betty.app import App
from betty.asyncio import sync
from betty.locale import Date, DateRange, Localized, Localey, Str
from betty.media_type import MediaType
from betty.model import Entity, get_entity_type_name, GeneratedEntityId
from betty.model.ancestry import Place, Person, PlaceName, Event, Described, HasLinks, HasCitations, Link, Dated, File, \
    Note, PersonName, HasMediaType, PresenceRole, Citation, Source, is_public, Presence, HasPrivacy, is_private, \
    HasNotes
from betty.string import upper_camel_case_to_lower_camel_case

T = TypeVar('T')


def validate(data: Any, schema_definition: str, app: App) -> None:
    with open(path.join(path.dirname(__file__), 'assets', 'public', 'static', 'schema.json'), encoding='utf-8') as f:
        json_data = f.read()
    schema = stdjson.loads(json_data)
    # @todo Can we set the schema ID somehow without making the entire JSON schema file a Jinja2 template?
    schema_id = app.static_url_generator.generate('schema.json', absolute=True)
    schema['$id'] = schema_id
    ref_resolver = RefResolver(schema_id, schema)
    jsonschema.validate(
        data, schema['definitions'][schema_definition], resolver=ref_resolver)


class JSONEncoder(stdjson.JSONEncoder):
    def __init__(
        self,
        app: App,
        *args: Any,
        **kwargs: Any,
    ):
        stdjson.JSONEncoder.__init__(self, *args, **kwargs)
        self._app = app
        self._mappers: dict[type[Any], Callable[[Any], Any]] = {
            Str: str,
            Path: str,
            PlaceName: self._encode_localized_name,
            Place: self._encode_place,
            Point: self._encode_coordinates,
            Person: self._encode_person,
            PersonName: self._encode_person_name,
            File: self._encode_file,
            Event: self._encode_event,
            PresenceRole: self._encode_presence_role,
            Date: self._encode_date,
            DateRange: self._encode_date_range,
            Citation: self._encode_citation,
            Source: self._encode_source,
            Link: self._encode_link,
            Note: self._encode_note,
            MediaType: self._encode_media_type,
        }

    def default(self, o: type[T]) -> Callable[[T], Any]:
        for mapper_type in self._mappers:
            if isinstance(o, mapper_type):
                return cast(Callable[[T], Any], self._mappers[mapper_type](o))
        return stdjson.JSONEncoder.default(self, o)  # type: ignore[no-any-return]

    def _generate_url(self, resource: Any, media_type='application/json', locale: Localey | None = None) -> str:
        return self._app.url_generator.generate(resource, media_type, locale=locale)

    def _encode_schema(self, encoded: dict[str, Any], defintion: str) -> None:
        encoded['$schema'] = self._app.static_url_generator.generate(
            'schema.json#/definitions/%s' % defintion)

    @sync
    async def _encode_entity(self, encoded: dict[str, Any], entity: Entity) -> None:
        self._encode_schema(encoded, upper_camel_case_to_lower_camel_case(get_entity_type_name(entity)))

        if 'links' not in encoded:
            encoded['links'] = []

        if '@context' not in encoded:
            encoded['@context'] = {}

        if not isinstance(entity.id, GeneratedEntityId):
            encoded['id'] = entity.id

            canonical = Link(
                self._generate_url(entity),
                relationship='canonical',
                media_type=MediaType('application/json'),
            )
            encoded['links'].append(canonical)

            if is_public(entity):
                for locale in self._app.project.configuration.locales:
                    localized_html_url = self._generate_url(entity, media_type='text/html', locale=locale)
                    localized_html_link = Link(
                        localized_html_url,
                        relationship='alternate',
                        media_type=MediaType('text/html'),
                        locale=locale,
                    )
                    encoded['links'].append(localized_html_link)

        if isinstance(entity, HasPrivacy):
            encoded['private'] = is_private(entity)
        if isinstance(entity, Dated):
            self._encode_dated(encoded, entity)
        if isinstance(entity, HasCitations):
            self._encode_has_citations(encoded, entity)
        if isinstance(entity, HasLinks):
            self._encode_has_links(encoded, entity)
        if isinstance(entity, HasNotes):
            self._encode_has_notes(encoded, entity)

    def _encode_described(self, encoded: dict[str, Any], described: Described) -> None:
        if described.description is not None:
            encoded['description'] = described.description
            encoded.update({
                '@context': {},
            })
            encoded['@context']['description'] = 'https://schema.org/description'

    def _encode_dated(self, encoded: dict[str, Any], dated: Dated) -> None:
        if is_public(dated):
            if dated.date is not None:
                encoded['date'] = dated.date

    def _encode_date(self, date: Date) -> dict[str, Any]:
        encoded: dict[str, Any] = {}
        if date.year:
            encoded['year'] = date.year
        if date.month:
            encoded['month'] = date.month
        if date.day:
            encoded['day'] = date.day
        return encoded

    def _encode_date_range(self, date: DateRange) -> dict[str, Any]:
        encoded: dict[str, Any] = {}
        if date.start:
            encoded['start'] = date.start
        if date.end:
            encoded['end'] = date.end
        return encoded

    def _encode_localized(self, encoded: dict[str, Any], localized: Localized) -> None:
        if localized.locale is not None:
            encoded['locale'] = localized.locale

    def _encode_has_media_type(self, encoded: dict[str, Any], media: HasMediaType) -> None:
        if is_public(media):
            if media.media_type is not None:
                encoded['mediaType'] = media.media_type

    def _encode_has_links(self, encoded: dict[str, Any], has_links: HasLinks) -> None:
        if 'links' not in encoded:
            encoded['links'] = []
        if is_public(has_links):
            for link in has_links.links:
                encoded['links'].append(link)

    def _encode_link(self, link: Link) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            'url': link.url,
        }
        if link.label is not None:
            encoded['label'] = link.label
        if link.relationship is not None:
            encoded['relationship'] = link.relationship
        self._encode_localized(encoded, link)
        self._encode_has_media_type(encoded, link)
        return encoded

    def _encode_has_citations(self, encoded: dict[str, Any], has_citations: HasCitations) -> None:
        encoded['citations'] = [
            self._generate_url(citation)
            for citation
            in has_citations.citations
            if not isinstance(citation.id, GeneratedEntityId)
        ]

    def _encode_coordinates(self, coordinates: Point) -> dict[str, Any]:
        return {
            '@context': {
                'latitude': 'https://schema.org/latitude',
                'longitude': 'https://schema.org/longitude',
            },
            '@type': 'https://schema.org/GeoCoordinates',
            'latitude': coordinates.latitude,
            'longitude': coordinates.longitude,
        }

    def _encode_localized_name(self, name: PlaceName) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            'name': name.name,
        }
        self._encode_localized(encoded, name)
        return encoded

    def _encode_place(self, place: Place) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            '@context': {
                'events': 'https://schema.org/event',
                'enclosedBy': 'https://schema.org/containedInPlace',
                'encloses': 'https://schema.org/containsPlace',
            },
            '@type': 'https://schema.org/Place',
            'names': place.names,
            'events': [
                self._generate_url(event)
                for event
                in place.events
                if not isinstance(event.id, GeneratedEntityId)
            ],
            'enclosedBy': [
                self._generate_url(enclosure.enclosed_by)
                for enclosure
                in place.enclosed_by
                if enclosure.enclosed_by is not None and not isinstance(enclosure.enclosed_by.id, GeneratedEntityId)
            ],
            'encloses': [
                self._generate_url(enclosure.encloses)
                for enclosure
                in place.encloses
                if enclosure.encloses is not None and not isinstance(enclosure.encloses.id, GeneratedEntityId)
            ],
        }
        self._encode_entity(encoded, place)
        if place.coordinates is not None:
            encoded['coordinates'] = place.coordinates
            encoded['@context']['coordinates'] = 'https://schema.org/geo'
        return encoded

    def _encode_person(self, person: Person) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            '@context': {
                'parents': 'https://schema.org/parent',
                'children': 'https://schema.org/child',
                'siblings': 'https://schema.org/sibling',
            },
            '@type': 'https://schema.org/Person',
            'names': [],
            'parents': [
                self._generate_url(parent)
                for parent
                in person.parents
                if not isinstance(parent.id, GeneratedEntityId)
            ],
            'children': [
                self._generate_url(child)
                for child
                in person.children
                if not isinstance(child.id, GeneratedEntityId)
            ],
            'siblings': [
                self._generate_url(sibling)
                for sibling
                in person.siblings
                if not isinstance(sibling.id, GeneratedEntityId)
            ],
            'presences': [
                self._encode_person_presence(presence)
                for presence
                in person.presences
                if presence.event is not None and not isinstance(presence.event.id, GeneratedEntityId)
            ],
        }
        if person.public:
            for name in person.names:
                if name.public:
                    encoded['names'].append(name)
        self._encode_entity(encoded, person)
        return encoded

    def _encode_person_presence(self, presence: Presence) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            '@context': {
                'event': 'https://schema.org/performerIn',
            },
            'event': self._generate_url(presence.event),
        }
        if is_public(presence.person):
            encoded['role'] = presence.role
        return encoded

    def _encode_person_name(self, name: PersonName) -> dict[str, Any]:
        encoded: dict[str, Any] = {}
        if name.public:
            if name.individual is not None or name.affiliation is not None:
                encoded.update({
                    '@context': {},
                })
            if name.individual is not None:
                encoded['@context']['individual'] = 'https://schema.org/givenName'
                encoded['individual'] = name.individual
            if name.affiliation is not None:
                encoded['@context']['affiliation'] = 'https://schema.org/familyName'
                encoded['affiliation'] = name.affiliation
        return encoded

    def _encode_file(self, file: File) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            'entities': [
                self._generate_url(entity)
                for entity
                in file.entities
                if not isinstance(entity.id, GeneratedEntityId)
            ],
        }
        self._encode_entity(encoded, file)
        self._encode_has_media_type(encoded, file)
        return encoded

    def _encode_event(self, event: Event) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            '@type': 'https://schema.org/Event',
            'type': event.event_type.name(),
            'presences': [
                self._encode_event_presence(presence)
                for presence
                in event.presences
                if presence.person is not None and not isinstance(presence.person.id, GeneratedEntityId)
            ],
        }
        self._encode_entity(encoded, event)
        if event.place is not None and not isinstance(event.place.id, GeneratedEntityId):
            encoded['place'] = self._generate_url(event.place)
            encoded.update({
                '@context': {},
            })
            encoded['@context']['place'] = 'https://schema.org/location'
        return encoded

    def _encode_event_presence(self, presence: Presence) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            '@context': {
                'person': 'https://schema.org/actor',
            },
            'person': self._generate_url(presence.person),
        }
        if is_public(presence.person):
            encoded['role'] = presence.role
        return encoded

    def _encode_presence_role(self, role: PresenceRole) -> str:
        return role.name()

    def _encode_citation(self, citation: Citation) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            '@type': 'https://schema.org/Thing',
            'facts': []
        }
        if citation.source is not None and not isinstance(citation.source.id, GeneratedEntityId):
            encoded['source'] = self._generate_url(citation.source)
        for fact in citation.facts:
            if isinstance(fact.id, GeneratedEntityId):
                continue
            encoded['facts'].append(
                self._generate_url(fact),
            )
        self._encode_entity(encoded, citation)
        return encoded

    def _encode_source(self, source: Source) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            '@type': 'https://schema.org/Thing',
            'contains': [
                self._generate_url(contained)
                for contained
                in source.contains
                if not isinstance(contained.id, GeneratedEntityId)
            ],
            'citations': [
                self._generate_url(citation)
                for citation
                in source.citations
                if not isinstance(citation.id, GeneratedEntityId)
            ],
        }
        if source.contained_by is not None and not isinstance(source.contained_by.id, GeneratedEntityId):
            encoded['containedBy'] = self._generate_url(source.contained_by)
        self._encode_entity(encoded, source)
        if is_public(source):
            if source.name is not None:
                encoded['@context']['name'] = 'https://schema.org/name'
                encoded['name'] = source.name
            if source.author is not None:
                encoded['author'] = source.author
            if source.publisher is not None:
                encoded['publisher'] = source.publisher
        return encoded

    def _encode_has_notes(self, encoded: dict[str, Any], has_notes: HasNotes) -> None:
        encoded['notes'] = [
            self._generate_url(note)
            for note
            in has_notes.notes
            if not isinstance(note.id, GeneratedEntityId)
        ]

    def _encode_note(self, note: Note) -> dict[str, Any]:
        encoded: dict[str, Any] = {
            '@type': 'https://schema.org/Thing',
        }
        self._encode_entity(encoded, note)
        if is_public(note):
            encoded['text'] = note.text
        return encoded

    def _encode_media_type(self, media_type: MediaType) -> str:
        return str(media_type)



from __future__ import annotations

import functools
import traceback
from types import TracebackType
from typing import Callable, Any, TypeVar, Generic, TYPE_CHECKING, ParamSpec

from PyQt6.QtCore import QMetaObject, Qt, Q_ARG, QObject
from PyQt6.QtGui import QCloseEvent, QIcon
from PyQt6.QtWidgets import QWidget, QMessageBox

from betty.app import App
from betty.gui.locale import LocalizedMessageBox

if TYPE_CHECKING:
    from betty.gui import QWidgetT


T = TypeVar('T')
P = ParamSpec('P')


class _ExceptionCatcher(Generic[P, T]):
    def __init__(
            self,
            f: Callable[P, T] | None = None,
            parent: QWidget | None = None,
            close_parent: bool = False,
            instance: QWidget | None = None,
    ):
        if f:
            functools.update_wrapper(self, f)
        self._f = f
        if close_parent and not parent:
            raise ValueError('No parent was given to close.')
        self._parent = instance if parent is None else parent
        self._close_parent = close_parent
        self._instance = instance

    def __get__(self, instance: QWidgetT | None, owner: type[QWidgetT] | None = None) -> Any:
        if instance is None:
            return self
        assert isinstance(instance, QWidget)
        return type(self)(self._f, instance, self._close_parent, instance)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if not self._f:
            raise RuntimeError('This exception catcher is not callable, but you can use it as a context manager instead using a `with` statement.')
        if self._instance is not None:
            args = (self._instance, *args)  # type: ignore[assignment]
        with self:
            return self._f(*args, **kwargs)

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> bool | None:
        from betty.gui import BettyApplication

        if exc_val is None:
            return None
        QMetaObject.invokeMethod(
            BettyApplication.instance(),
            '_catch_exception',
            Qt.ConnectionType.QueuedConnection,
            Q_ARG(Exception, exc_val),
            Q_ARG(QObject, self._parent),
            Q_ARG(bool, self._close_parent),
        )
        return True


# Alias the class so its original name follows the PEP code style, but the alias follows the decorator code style.
catch_exceptions = _ExceptionCatcher


class Error(LocalizedMessageBox):
    def __init__(
        self,
        app: App,
        message: str,
        *args: Any,
        close_parent: bool = False,
        **kwargs: Any,
    ):
        super().__init__(app, *args, **kwargs)
        self._close_parent = close_parent
        self.setWindowTitle('{error} - Betty'.format(error=self._app.localizer._("Error")))
        self.setText(message)

        standard_button_type = QMessageBox.StandardButton.Close
        self.setStandardButtons(standard_button_type)
        close_button = self.button(standard_button_type)
        assert close_button is not None
        close_button.setIcon(QIcon())
        self.setDefaultButton(standard_button_type)
        self.setEscapeButton(standard_button_type)
        close_button.clicked.connect(self.close)

    def closeEvent(self, a0: QCloseEvent | None) -> None:
        if self._close_parent:
            parent = self.parent()
            if isinstance(parent, QWidget):
                parent.close()
        super().closeEvent(a0)


ErrorT = TypeVar('ErrorT', bound=Error)


class ExceptionError(Error):
    def __init__(self, app: App, exception: Exception, *args: Any, **kwargs: Any):
        super().__init__(app, str(exception), *args, **kwargs)
        self.exception = exception


class UnexpectedExceptionError(ExceptionError):
    def __init__(self, app: App, exception: Exception, *args: Any, **kwargs: Any):
        super().__init__(app, exception, *args, **kwargs)
        self.setText(self._app.localizer._('An unexpected error occurred and Betty could not complete the task. Please <a href="{report_url}">report this problem</a> and include the following details, so the team behind Betty can address it.').format(
            report_url='https://github.com/bartfeenstra/betty/issues',
        ))
        self.setTextFormat(Qt.TextFormat.RichText)
        self.setDetailedText(''.join(traceback.format_exception(type(exception), exception, exception.__traceback__)))



from __future__ import annotations

from contextlib import contextmanager
from textwrap import indent
from typing import Iterator, Self

from betty.error import UserFacingError
from betty.locale import Localizable, Localizer, Str


class SerdeError(UserFacingError, ValueError):
    """
    A serialization or deserialization error.
    """

    def __init__(self, message: Localizable):
        super().__init__(message)
        self._contexts: tuple[Localizable, ...] = ()

    def localize(self, localizer: Localizer) -> str:
        localized_contexts = map(lambda context: context.localize(localizer), self._contexts)
        return (super().localize(localizer) + '\n' + indent('\n'.join(localized_contexts), '- ')).strip()

    def raised(self, error_type: type[SerdeError]) -> bool:
        return isinstance(self, error_type)

    @property
    def contexts(self) -> tuple[Localizable, ...]:
        return self._contexts

    def with_context(self, *contexts: Localizable) -> Self:
        """
        Add a message describing the error's context.
        """
        self_copy = self._copy()
        self_copy._contexts = (*self._contexts, *contexts)
        return self_copy

    def _copy(self) -> Self:
        return type(self)(self._localizable_message)


class SerdeErrorCollection(SerdeError):
    """
    A collection of zero or more serialization or deserialization errors.
    """

    def __init__(
        self,
        errors: list[SerdeError] | None = None,
    ):
        super().__init__(Str._('The following errors occurred'))
        self._errors: list[SerdeError] = errors or []

    def __iter__(self) -> Iterator[SerdeError]:
        yield from self._errors

    def localize(self, localizer: Localizer) -> str:
        return '\n\n'.join(map(lambda error: error.localize(localizer), self._errors))

    def __reduce__(self) -> tuple[type[Self], tuple[list[SerdeError]]]:  # type: ignore[override]
        return type(self), (self._errors,)

    def __len__(self) -> int:
        return len(self._errors)

    def raised(self, error_type: type[SerdeError]) -> bool:
        for error in self._errors:
            if error.raised(error_type):
                return True
        return False

    @property
    def valid(self) -> bool:
        return len(self._errors) == 0

    @property
    def invalid(self) -> bool:
        return not self.valid

    @contextmanager
    def assert_valid(self) -> Iterator[Self]:
        if self.invalid:
            raise self
        with self.catch():
            yield self
        if self.invalid:  # type: ignore[redundant-expr]
            raise self

    def append(self, *errors: SerdeError) -> None:
        for error in errors:
            if isinstance(error, SerdeErrorCollection):
                self.append(*error)
            else:
                self._errors.append(error.with_context(*self._contexts))

    def with_context(self, *contexts: Localizable) -> Self:
        self_copy = super().with_context(*contexts)
        self_copy._errors = [error.with_context(*contexts) for error in self._errors]
        return self_copy

    def _copy(self) -> Self:
        return type(self)()

    @contextmanager
    def catch(self, *contexts: Localizable) -> Iterator[SerdeErrorCollection]:
        context_errors: SerdeErrorCollection = SerdeErrorCollection()
        if contexts:
            context_errors = context_errors.with_context(*contexts)
        try:
            yield context_errors
        except SerdeError as e:
            context_errors.append(e)
        self.append(*context_errors)



from __future__ import annotations

import hashlib
import os
from collections import deque
from contextlib import suppress
from os.path import getmtime
from pathlib import Path
from shutil import copy2
from types import TracebackType
from typing import AsyncIterable, AsyncContextManager, Sequence

import aiofiles
from aiofiles.os import makedirs
from aiofiles.ospath import exists
from aiofiles.threadpool.text import AsyncTextIOWrapper

from betty import _ROOT_DIRECTORY_PATH

ROOT_DIRECTORY_PATH = _ROOT_DIRECTORY_PATH


ASSETS_DIRECTORY_PATH = ROOT_DIRECTORY_PATH / 'betty' / 'assets'


HOME_DIRECTORY_PATH = Path.home() / '.betty'


CACHE_DIRECTORY_PATH = HOME_DIRECTORY_PATH / 'cache'


async def iterfiles(path: Path) -> AsyncIterable[Path]:
    for dir_path, _, filenames in os.walk(str(path)):
        for filename in filenames:
            yield Path(dir_path) / filename


def hashfile(path: Path) -> str:
    return hashlib.md5(':'.join([str(getmtime(path)), str(path)]).encode('utf-8')).hexdigest()


class FileSystem:
    class _Open:
        def __init__(self, fs: FileSystem, file_paths: tuple[Path, ...]):
            self._fs = fs
            self._file_paths = file_paths
            self._file: AsyncContextManager[AsyncTextIOWrapper] | None = None

        async def __aenter__(self) -> AsyncTextIOWrapper:
            for file_path in map(Path, self._file_paths):
                for fs_path, fs_encoding in self._fs._paths:
                    with suppress(FileNotFoundError):
                        self._file = aiofiles.open(fs_path / file_path, encoding=fs_encoding)
                        return await self._file.__aenter__()
            raise FileNotFoundError

        async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
            if self._file is not None:
                await self._file.__aexit__(None, None, None)

    def __init__(self, *paths: tuple[Path, str | None]):
        self._paths = deque(paths)

    def __len__(self) -> int:
        return len(self._paths)

    @property
    def paths(self) -> Sequence[tuple[Path, str | None]]:
        return list(self._paths)

    def prepend(self, path: Path, fs_encoding: str | None = None) -> None:
        self._paths.appendleft((path, fs_encoding))

    def clear(self) -> None:
        self._paths.clear()

    def open(self, *file_paths: Path) -> _Open:
        return self._Open(self, file_paths)

    async def copy2(self, source_path: Path, destination_path: Path) -> Path:
        for fs_path, _ in self._paths:
            with suppress(FileNotFoundError):
                copy2(fs_path / source_path, destination_path)
                return destination_path
        tried_paths = [str(fs_path / source_path) for fs_path, _ in self._paths]
        raise FileNotFoundError('Could not find any of %s.' % ', '.join(tried_paths))

    async def copytree(self, source_path: Path, destination_path: Path) -> AsyncIterable[Path]:
        destination_paths = set()
        for fs_path, _ in self._paths:
            async for file_source_path in iterfiles(fs_path / source_path):
                file_destination_path = destination_path / file_source_path.relative_to(fs_path / source_path)
                if not await exists(file_destination_path):
                    await makedirs(file_destination_path.parent, exist_ok=True)
                    copy2(file_source_path, file_destination_path)
                    if file_destination_path not in destination_paths:
                        destination_paths.add(file_destination_path)
                        yield file_destination_path



from __future__ import annotations

import json
import logging
import multiprocessing
import os
import queue
import shutil
import threading
from concurrent.futures import ProcessPoolExecutor, Executor, Future, as_completed
from contextlib import suppress
from ctypes import c_char_p
from multiprocessing.managers import ValueProxy
from pathlib import Path
from types import TracebackType
from typing import cast, AsyncContextManager, Self, Any, ParamSpec, Callable, Concatenate, MutableSequence

import aiofiles
import dill
from aiofiles import os as aiofiles_os
from aiofiles.os import makedirs
from aiofiles.threadpool.text import AsyncTextIOWrapper

from betty.app import App
from betty.asyncio import sync, gather
from betty.locale import get_display_name
from betty.model import get_entity_type_name, UserFacingEntity, Entity, GeneratedEntityId
from betty.model.ancestry import is_public
from betty.openapi import Specification
from betty.serde.dump import DictDump, Dump
from betty.string import camel_case_to_kebab_case, camel_case_to_snake_case
from betty.task import Context

_GenerationProcessPoolTaskP = ParamSpec('_GenerationProcessPoolTaskP')


def getLogger() -> logging.Logger:
    return logging.getLogger(__name__)


class Generator:
    async def generate(self, task_context: GenerationContext) -> None:
        raise NotImplementedError(repr(self))


class GenerationContext(Context):
    def __init__(self, app: App):
        super().__init__()
        self._pickled_app = multiprocessing.Manager().Value(c_char_p, dill.dumps(app))
        self._unpickle_app_lock: threading.Lock = multiprocessing.Manager().Lock()
        self._app: App | None = None

    def __getstate__(self) -> tuple[threading.Lock, MutableSequence[str], ValueProxy[bytes]]:
        return self._claims_lock, self._claimed_task_ids, self._pickled_app

    def __setstate__(self, state: tuple[threading.Lock, MutableSequence[str], ValueProxy[bytes]]) -> None:
        self._claims_lock, self._claimed_task_ids, self._pickled_app = state
        self._unpickle_app_lock = multiprocessing.Manager().Lock()
        self._app = None

    @property
    def app(self) -> App:
        with self._unpickle_app_lock:
            if self._app is None:
                self._app = cast(App, dill.loads(self._pickled_app.value))
        return self._app


class _GenerationProcessPool:
    def __init__(self, app: App, task_context: GenerationContext):
        self._app = app
        self._task_context = task_context
        self._queue = multiprocessing.Manager().Queue()
        self._cancel = multiprocessing.Manager().Event()
        self._finish = multiprocessing.Manager().Event()
        self._executor: Executor | None = None
        self._workers: list[Future[None]] = []

    async def __aenter__(self) -> Self:
        self._executor = ProcessPoolExecutor(max_workers=self._app.concurrency)
        for _ in range(0, self._app.concurrency):
            self._workers.append(self._executor.submit(_GenerationProcessPoolWorker(
                self._queue,
                self._cancel,
                self._finish,
                self._app.concurrency,
                self._task_context,
            )))
        return self

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        assert self._executor is not None
        if exc_val is None:
            self._finish.set()
        else:
            self._cancel.set()
        try:
            for worker in as_completed(self._workers):
                worker.result()
        except BaseException:
            self._cancel.set()
            raise
        finally:
            self._executor.shutdown()

    def delegate(
        self,
        task_callable: Callable[Concatenate[GenerationContext, _GenerationProcessPoolTaskP], Any],
        *task_args: _GenerationProcessPoolTaskP.args,
        **task_kwargs: _GenerationProcessPoolTaskP.kwargs,
    ) -> None:
        self._queue.put((task_callable, task_args, task_kwargs))


class _GenerationProcessPoolWorker:
    def __init__(
        self,
        task_queue: queue.Queue[tuple[Callable[Concatenate[GenerationContext, _GenerationProcessPoolTaskP], Any], _GenerationProcessPoolTaskP.args, _GenerationProcessPoolTaskP.kwargs]],
        cancel: threading.Event,
        finish: threading.Event,
        async_concurrency: int,
        task_context: GenerationContext,
    ):
        self._task_queue = task_queue
        self._cancel = cancel
        self._finish = finish
        self._async_concurrency = async_concurrency
        self._context = task_context

    @sync
    async def __call__(self) -> None:
        async with self._context.app:
            await gather(*(
                self._perform_tasks()
                for _ in range(0, self._async_concurrency)
            ))

    async def _perform_tasks(self) -> None:
        while not self._cancel.is_set():
            try:
                task_callable, task_args, task_kwargs = self._task_queue.get_nowait()
            except queue.Empty:
                if self._finish.is_set():
                    return
            else:
                await task_callable(
                    self._context,
                    *task_args,
                    **task_kwargs,
                )


async def generate(app: App) -> None:
    logger = getLogger()
    task_context = GenerationContext(app)

    with suppress(FileNotFoundError):
        shutil.rmtree(app.project.configuration.output_directory_path)
    await aiofiles_os.makedirs(app.project.configuration.output_directory_path, exist_ok=True)
    logger.info(app.localizer._('Generating your site to {output_directory}.').format(output_directory=app.project.configuration.output_directory_path))

    # The static public assets may be overridden depending on the number of locales rendered, so ensure they are
    # generated before anything else.
    await _generate_static_public(app, task_context)

    locales = app.project.configuration.locales

    async with _GenerationProcessPool(app, task_context) as process_pool:
        process_pool.delegate(_generate_dispatch)
        process_pool.delegate(_generate_openapi)

        for locale in locales:
            process_pool.delegate(_generate_public, locale)

        for entity_type in app.entity_types:
            if not issubclass(entity_type, UserFacingEntity):
                continue
            if app.project.configuration.entity_types[entity_type].generate_html_list:
                for locale in locales:
                    process_pool.delegate(_generate_entity_type_list_html, locale, entity_type)
            process_pool.delegate(_generate_entity_type_list_json, entity_type)
            for entity in app.project.ancestry[entity_type]:
                if isinstance(entity.id, GeneratedEntityId):
                    continue

                process_pool.delegate(_generate_entity_json, entity_type, entity.id)
                if is_public(entity):
                    for locale in locales:
                        process_pool.delegate(_generate_entity_html, locale, entity_type, entity.id)

    # Log the generated pages.
    for locale in app.project.configuration.locales:
        locale_label = get_display_name(locale, app.localizer.locale)
        for entity_type in app.entity_types:
            if issubclass(entity_type, UserFacingEntity):
                logger.info(app.localizer._('Generated pages for {count} {entity_type} in {locale}.').format(
                    count=len(app.project.ancestry[entity_type]),
                    entity_type=entity_type.entity_type_label_plural().localize(app.localizer),
                    locale=locale_label,
                ))

    os.chmod(app.project.configuration.output_directory_path, 0o755)
    for directory_path_str, subdirectory_names, file_names in os.walk(app.project.configuration.output_directory_path):
        directory_path = Path(directory_path_str)
        for subdirectory_name in subdirectory_names:
            os.chmod(directory_path / subdirectory_name, 0o755)
        for file_name in file_names:
            os.chmod(directory_path / file_name, 0o644)


async def create_file(path: Path) -> AsyncContextManager[AsyncTextIOWrapper]:
    await makedirs(path.parent, exist_ok=True)
    return cast(AsyncContextManager[AsyncTextIOWrapper], aiofiles.open(path, 'w', encoding='utf-8'))


async def create_html_resource(path: Path) -> AsyncContextManager[AsyncTextIOWrapper]:
    return await create_file(path / 'index.html')


async def create_json_resource(path: Path) -> AsyncContextManager[AsyncTextIOWrapper]:
    return await create_file(path / 'index.json')


async def _generate_dispatch(
    task_context: GenerationContext,
) -> None:
    app = task_context.app
    await app.dispatcher.dispatch(Generator)(task_context),


async def _generate_public(
    task_context: GenerationContext,
    locale: str,
) -> None:
    app = task_context.app
    locale_label = get_display_name(locale, app.localizer.locale)
    getLogger().info(app.localizer._('Generating localized public files in {locale}...').format(
        locale=locale_label,
    ))
    async for file_path in app.assets.copytree(Path('public') / 'localized', app.project.configuration.localize_www_directory_path(locale)):
        await app.renderer.render_file(
            file_path,
            task_context=task_context,
            localizer=app.localizers[locale],
        )


async def _generate_static_public(
    app: App,
    task_context: Context,
) -> None:
    getLogger().info(app.localizer._('Generating static public files...'))
    async for file_path in app.assets.copytree(Path('public') / 'static', app.project.configuration.www_directory_path):
        await app.renderer.render_file(
            file_path,
            task_context=task_context,
        )


async def _generate_entity_type_list_html(
    task_context: GenerationContext,
    locale: str,
    entity_type: type[Entity],
) -> None:
    app = task_context.app
    entity_type_name_fs = camel_case_to_kebab_case(get_entity_type_name(entity_type))
    entity_type_path = app.project.configuration.localize_www_directory_path(locale) / entity_type_name_fs
    template = app.jinja2_environment.negotiate_template([
        f'entity/page-list--{entity_type_name_fs}.html.j2',
        'entity/page-list.html.j2',
    ])
    rendered_html = await template.render_async(
        task_context=task_context,
        localizer=app.localizers[locale],
        page_resource=f'/{entity_type_name_fs}/index.html',
        entity_type=entity_type,
        entities=app.project.ancestry[entity_type],
    )
    async with await create_html_resource(entity_type_path) as f:
        await f.write(rendered_html)
    locale_label = get_display_name(locale, app.localizer.locale)
    getLogger().info(app.localizer._('Generated the listing page for {entity_type} in {locale}.').format(
        entity_type=entity_type.entity_type_label_plural().localize(app.localizer),
        locale=locale_label,
    ))


async def _generate_entity_type_list_json(
    task_context: GenerationContext,
    entity_type: type[Entity],
) -> None:
    app = task_context.app
    entity_type_name = get_entity_type_name(entity_type)
    entity_type_name_fs = camel_case_to_kebab_case(get_entity_type_name(entity_type))
    entity_type_path = app.project.configuration.www_directory_path / entity_type_name_fs
    data: DictDump[Dump] = {
        '$schema': app.static_url_generator.generate('schema.json#/definitions/%sCollection' % entity_type_name, absolute=True),
        'collection': []
    }
    for entity in app.project.ancestry[entity_type]:
        cast(list[str], data['collection']).append(
            app.url_generator.generate(
                entity,
                'application/json',
                absolute=True,
            ))
    rendered_json = json.dumps(data)
    async with await create_json_resource(entity_type_path) as f:
        await f.write(rendered_json)


async def _generate_entity_html(
    task_context: GenerationContext,
    locale: str,
    entity_type: type[Entity],
    entity_id: str,
) -> None:
    app = task_context.app
    entity = app.project.ancestry[entity_type][entity_id]
    entity_type_name_fs = camel_case_to_kebab_case(get_entity_type_name(entity))
    entity_path = app.project.configuration.localize_www_directory_path(locale) / entity_type_name_fs / entity.id
    rendered_html = await app.jinja2_environment.negotiate_template([
        f'entity/page--{entity_type_name_fs}.html.j2',
        'entity/page.html.j2',
    ]).render_async(
        task_context=task_context,
        localizer=app.localizers[locale],
        page_resource=entity,
        entity_type=entity.type,
        entity=entity,
    )
    async with await create_html_resource(entity_path) as f:
        await f.write(rendered_html)


async def _generate_entity_json(
    task_context: GenerationContext,
    entity_type: type[Entity],
    entity_id: str,
) -> None:
    app = task_context.app
    entity_type_name_fs = camel_case_to_kebab_case(get_entity_type_name(entity_type))
    entity_path = app.project.configuration.www_directory_path / entity_type_name_fs / entity_id
    rendered_json = json.dumps(app.project.ancestry[entity_type][entity_id], cls=app.json_encoder)
    async with await create_json_resource(entity_path) as f:
        await f.write(rendered_json)


async def _generate_openapi(
    task_context: GenerationContext,
) -> None:
    app = task_context.app
    getLogger().info(app.localizer._('Generating OpenAPI specification...'))
    api_directory_path = app.project.configuration.www_directory_path / 'api'
    rendered_json = json.dumps(Specification(app).build())
    async with await create_json_resource(api_directory_path) as f:
        await f.write(rendered_json)


def _get_entity_type_jinja2_name(entity_type_name: str) -> str:
    return camel_case_to_snake_case(entity_type_name).replace('.', '__')



from __future__ import annotations

import multiprocessing
from collections.abc import MutableSequence


class Context:
    def __init__(self):
        self._claims_lock = multiprocessing.Manager().Lock()
        self._claimed_task_ids: MutableSequence[str] = multiprocessing.Manager().list()

    def claim(self, task_id: str) -> bool:
        with self._claims_lock:
            if task_id in self._claimed_task_ids:
                return False
            self._claimed_task_ids.append(task_id)
            return True



from __future__ import annotations

import os as stdos
import weakref
from contextlib import suppress
from graphlib import CycleError, TopologicalSorter
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Mapping, Self, final

import aiohttp
from reactives.instance import ReactiveInstance
from reactives.instance.property import reactive_property

from betty.app.extension import ListExtensions, Extension, Extensions, build_extension_type_graph, \
    CyclicDependencyError, ExtensionDispatcher, ConfigurableExtension, discover_extension_types
from betty.asyncio import sync, wait
from betty.cache import Cache
from betty.config import Configurable, FileBasedConfiguration
from betty.dispatch import Dispatcher
from betty.fs import FileSystem, ASSETS_DIRECTORY_PATH, HOME_DIRECTORY_PATH
from betty.locale import LocalizerRepository, get_data, DEFAULT_LOCALE, Localizer, Str
from betty.model import Entity, EntityTypeProvider
from betty.model.ancestry import Citation, Event, File, Person, PersonName, Presence, Place, Enclosure, \
    Source, Note
from betty.model.event_type import EventType, EventTypeProvider, Birth, Baptism, Adoption, Death, Funeral, Cremation, \
    Burial, Will, Engagement, Marriage, MarriageAnnouncement, Divorce, DivorceAnnouncement, Residence, Immigration, \
    Emigration, Occupation, Retirement, Correspondence, Confirmation
from betty.project import Project
from betty.render import Renderer, SequentialRenderer
from betty.serde.dump import minimize, void_none, Dump, VoidableDump
from betty.serde.load import AssertionFailed, Fields, Assertions, OptionalField, Asserter

if TYPE_CHECKING:
    from betty.jinja2 import Environment
    from betty.json import JSONEncoder
    from betty.serve import Server
    from betty.url import StaticUrlGenerator, ContentNegotiationUrlGenerator

CONFIGURATION_DIRECTORY_PATH = HOME_DIRECTORY_PATH / 'configuration'


class _AppExtensions(ListExtensions):
    def __init__(self):
        super().__init__([])

    def _update(self, extensions: list[list[Extension]]) -> None:
        self._extensions = extensions
        self.react.trigger()


class AppConfiguration(FileBasedConfiguration):
    def __init__(self):
        super().__init__()
        self._locale: str | None = None

    @property
    def configuration_file_path(self) -> Path:
        return CONFIGURATION_DIRECTORY_PATH / 'app.json'

    @configuration_file_path.setter
    def configuration_file_path(self, __) -> None:
        pass

    @configuration_file_path.deleter
    def configuration_file_path(self) -> None:
        pass

    @property
    @reactive_property
    def locale(self) -> str | None:
        return self._locale

    @locale.setter
    def locale(self, locale: str) -> None:
        try:
            get_data(locale)
        except ValueError:
            raise AssertionFailed(Str._(
                '"{locale}" is not a valid IETF BCP 47 language tag.',
                locale=locale,
            ))
        self._locale = locale

    def update(self, other: Self) -> None:
        self._locale = other._locale
        self.react.trigger()

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            OptionalField(
                'locale',
                Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'locale')),
        ),
        )(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return minimize({
            'locale': void_none(self.locale)
        }, True)


@final
class App(Configurable[AppConfiguration], ReactiveInstance):
    def __init__(
        self,
        configuration: AppConfiguration | None = None,
        project: Project | None = None,
    ):
        super().__init__()
        self._started = False
        self._configuration = configuration or AppConfiguration()
        self._assets: FileSystem | None = None
        self._extensions = _AppExtensions()
        self._extensions_initialized = False
        self._localization_initialized = False
        self._localizer: Localizer | None = None
        self._localizers: LocalizerRepository | None = None
        with suppress(FileNotFoundError):
            wait(self.configuration.read())
        self._project = project or Project()

        self._dispatcher: ExtensionDispatcher | None = None
        self._entity_types: set[type[Entity]] | None = None
        self._event_types: set[type[EventType]] | None = None
        self._url_generator: ContentNegotiationUrlGenerator | None = None
        self._static_url_generator: StaticUrlGenerator | None = None
        self._jinja2_environment: Environment | None = None
        self._renderer: Renderer | None = None
        self._http_client: aiohttp.ClientSession | None = None
        self._cache: Cache | None = None

    def __reduce__(self) -> tuple[
        type[App],
        tuple[
            AppConfiguration,
            Project,
        ],
    ]:
        return (
            App,
            (
                self._configuration,
                self._project,
            ),
        )

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        await self.stop()

    async def start(self) -> None:
        if self._started:
            raise RuntimeError('This app has started already.')
        self._started = True

    async def stop(self) -> None:
        self._started = False
        del self.http_client

    def __del__(self) -> None:
        if self._started:
            raise RuntimeError(f'{self} was started, but never stopped.')

    @property
    def project(self) -> Project:
        return self._project

    def discover_extension_types(self) -> set[type[Extension]]:
        return {*discover_extension_types(), *map(type, self._extensions.flatten())}

    @property
    def extensions(self) -> Extensions:
        if not self._extensions_initialized:
            self._extensions_initialized = True
            self._update_extensions()
            self.project.configuration.extensions.react(self._update_extensions)

        return self._extensions

    def _update_extensions(self) -> None:
        extension_types_enabled_in_configuration = set()
        for app_extension_configuration in self.project.configuration.extensions.values():
            if app_extension_configuration.enabled:
                app_extension_configuration.extension_type.enable_requirement().assert_met()
                extension_types_enabled_in_configuration.add(app_extension_configuration.extension_type)

        extension_types_sorter = TopologicalSorter(
            build_extension_type_graph(extension_types_enabled_in_configuration)
        )
        try:
            extension_types_sorter.prepare()
        except CycleError:
            raise CyclicDependencyError([
                app_extension_configuration.extension_type
                for app_extension_configuration
                in self.project.configuration.extensions.values()
            ])

        extensions = []
        while extension_types_sorter.is_active():
            extension_types_batch = extension_types_sorter.get_ready()
            extensions_batch = []
            for extension_type in extension_types_batch:
                if issubclass(extension_type, ConfigurableExtension) and extension_type in self.project.configuration.extensions:
                    extension: Extension = extension_type(self, configuration=self.project.configuration.extensions[extension_type].extension_configuration)
                else:
                    extension = extension_type(self)
                extensions_batch.append(extension)
                extension_types_sorter.done(extension_type)
            extensions.append(extensions_batch)
        self._extensions._update(extensions)

    @property
    @reactive_property(on_trigger_delete=True)
    def assets(self) -> FileSystem:
        if self._assets is None:
            self._assets = FileSystem()
            self._assets.prepend(ASSETS_DIRECTORY_PATH, 'utf-8')
            for extension in self.extensions.flatten():
                extension_assets_directory_path = extension.assets_directory_path()
                if extension_assets_directory_path is not None:
                    self._assets.prepend(extension_assets_directory_path, 'utf-8')
            self._assets.prepend(self.project.configuration.assets_directory_path)
        return self._assets

    @assets.deleter
    def assets(self) -> None:
        self._assets = None

    @property
    def dispatcher(self) -> Dispatcher:
        if self._dispatcher is None:
            self._dispatcher = ExtensionDispatcher(self.extensions)

        return self._dispatcher

    @property
    def url_generator(self) -> ContentNegotiationUrlGenerator:
        from betty.url import AppUrlGenerator

        if self._url_generator is None:
            self._url_generator = AppUrlGenerator(self)
        return self._url_generator

    @property
    def static_url_generator(self) -> StaticUrlGenerator:
        from betty.url import StaticPathUrlGenerator

        if self._static_url_generator is None:
            self._static_url_generator = StaticPathUrlGenerator(self.project.configuration)
        return self._static_url_generator

    @property
    def localizer(self) -> Localizer:
        """
        Get the application's localizer.

        The localizer MAY be out of sync with the locale set in the application configuration,
        if the locale is changed runtime. To keep almost every other part of the application
        simpler, changing the application locale requires an application restart for the changes
        to take effect.
        """
        if self._localizer is None:
            self._localizer = self.localizers.get_negotiated(self.configuration.locale or DEFAULT_LOCALE)
        return self._localizer

    @property
    def localizers(self) -> LocalizerRepository:
        if self._localizers is None:
            self._localizers = LocalizerRepository(self.assets)
        return self._localizers

    @property
    @reactive_property(on_trigger_delete=True)
    def jinja2_environment(self) -> Environment:
        if not self._jinja2_environment:
            from betty.jinja2 import Environment
            self._jinja2_environment = Environment(self)

        return self._jinja2_environment

    @jinja2_environment.deleter
    def jinja2_environment(self) -> None:
        self._jinja2_environment = None

    @property
    @reactive_property(on_trigger_delete=True)
    def renderer(self) -> Renderer:
        if not self._renderer:
            from betty.jinja2 import Jinja2Renderer

            self._renderer = SequentialRenderer([
                Jinja2Renderer(self.jinja2_environment, self.project.configuration),
            ])

        return self._renderer

    @renderer.deleter
    def renderer(self) -> None:
        self._renderer = None

    @property
    def concurrency(self) -> int:
        with suppress(KeyError):
            return int(stdos.environ['BETTY_CONCURRENCY'])
        # Assume that any machine that runs Betty has at least two CPU cores.
        return stdos.cpu_count() or 2

    @property
    def async_concurrency(self) -> int:
        return self.concurrency ** 2

    @property
    def json_encoder(self) -> type[JSONEncoder]:
        from betty.json import JSONEncoder
        return lambda *args, **kwargs: JSONEncoder(self)  # type: ignore[return-value]

    @property
    @reactive_property(on_trigger_delete=True)
    def http_client(self) -> aiohttp.ClientSession:
        if not self._http_client:
            self._http_client = aiohttp.ClientSession(connector=aiohttp.TCPConnector(limit_per_host=5))
            weakref.finalize(self, sync(self._http_client.close))
        return self._http_client

    @http_client.deleter
    @sync
    async def http_client(self) -> None:
        if self._http_client is not None:
            await self._http_client.close()
            self._http_client = None

    @property
    @reactive_property(on_trigger_delete=True)
    @sync
    async def entity_types(self) -> set[type[Entity]]:
        if self._entity_types is None:
            self._entity_types = set(await self.dispatcher.dispatch(EntityTypeProvider)()) | {
                Citation,
                Enclosure,
                Event,
                File,
                Note,
                Person,
                PersonName,
                Presence,
                Place,
                Source,
            }
        return self._entity_types

    @entity_types.deleter
    def entity_types(self) -> None:
        self._entity_types = None

    @property
    @reactive_property(on_trigger_delete=True)
    @sync
    async def event_types(self) -> set[type[EventType]]:
        if self._event_types is None:
            self._event_types = set(await self.dispatcher.dispatch(EventTypeProvider)()) | {
                Birth,
                Baptism,
                Adoption,
                Death,
                Funeral,
                Cremation,
                Burial,
                Will,
                Engagement,
                Marriage,
                MarriageAnnouncement,
                Divorce,
                DivorceAnnouncement,
                Residence,
                Immigration,
                Emigration,
                Occupation,
                Retirement,
                Correspondence,
                Confirmation,
            }
        return self._event_types

    @property
    def servers(self) -> Mapping[str, Server]:
        from betty import serve
        from betty.extension.demo import DemoServer

        return {
            server.name(): server
            for server
            in [
                *(
                    server
                    for extension in self.extensions.flatten()
                    if isinstance(extension, serve.ServerProvider)
                    for server in extension.servers
                ),
                serve.BuiltinServer(self),
                DemoServer(),
            ]
        }

    @property
    def cache(self) -> Cache:
        if self._cache is None:
            self._cache = Cache(self.localizer)
        return self._cache



from __future__ import annotations

from typing import Callable, cast, Iterator

from PyQt6.QtWidgets import QFormLayout, QLabel, QComboBox, QLineEdit, QWidget, QPushButton, QVBoxLayout, QHBoxLayout

from betty.app import App
from betty.gui.locale import LocalizedWidget
from betty.model import UserFacingEntity, Entity
from betty.project import EntityReference, EntityReferenceSequence


class EntityReferenceCollector(LocalizedWidget):
    def __init__(
        self,
        app: App,
        entity_reference: EntityReference[UserFacingEntity & Entity],
        label_builder: Callable[[], str] | None = None,
        caption_builder: Callable[[], str] | None = None,
    ):
        super().__init__(app)
        self._entity_reference = entity_reference
        self._label_builder = label_builder
        self._caption_builder = caption_builder

        self._layout = QFormLayout()
        self.setLayout(self._layout)

        if self._entity_reference.entity_type_is_constrained:
            self._entity_type_label = QLabel()
        else:
            def _update_entity_type() -> None:
                self._entity_reference.entity_type = self._entity_type.currentData()
            self._entity_type = QComboBox()
            self._entity_type.currentIndexChanged.connect(_update_entity_type)
            # @todo We use translated labels, and sort by them, but neither is reactive.
            entity_types = enumerate(sorted(cast(Iterator['UserFacingEntity & Entity'], filter(
                lambda entity_type: issubclass(entity_type, UserFacingEntity),
                self._app.entity_types,
            )), key=lambda entity_type: entity_type.entity_type_label().localize(self._app.localizer)))
            for i, entity_type in entity_types:
                self._entity_type.addItem(entity_type.entity_type_label().localize(self._app.localizer), entity_type)
                if entity_type == self._entity_reference.entity_type:
                    self._entity_type.setCurrentIndex(i)
            self._entity_type_label = QLabel()
            self._layout.addRow(self._entity_type_label, self._entity_type)

        def _update_entity_id() -> None:
            self._entity_reference.entity_id = self._entity_id.text()
        self._entity_id = QLineEdit()
        self._entity_id.textChanged.connect(_update_entity_id)
        self._entity_id_label = QLabel()
        self._layout.addRow(self._entity_id_label, self._entity_id)

        self._set_translatables()

    def _do_set_translatables(self) -> None:
        if self._entity_reference.entity_type:
            self._entity_id_label.setText(self._app.localizer._('{entity_type_label} ID').format(
                entity_type_label=self._entity_reference.entity_type.entity_type_label().localize(self._app.localizer),
            ))
        else:
            self._entity_id_label.setText(self._app.localizer._('Entity ID'))


class EntityReferenceSequenceCollector(LocalizedWidget):
    def __init__(
        self,
        app: App,
        entity_references: EntityReferenceSequence[UserFacingEntity & Entity],
        label_builder: Callable[[], str] | None = None,
        caption_builder: Callable[[], str] | None = None,
    ):
        super().__init__(app)
        self._entity_references = entity_references
        self._label_builder = label_builder
        self._caption_builder = caption_builder
        self._entity_reference_collectors: list[EntityReferenceCollector] = [
            EntityReferenceCollector(self._app, entity_reference)
            for entity_reference
            in entity_references
        ]

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        if label_builder:
            self._label = QLabel()
            self._layout.addWidget(self._label)

        self._entity_reference_collectors_widget = QWidget()
        self._entity_reference_collectors_layout = QVBoxLayout()
        self._entity_reference_collectors_widget.setLayout(self._entity_reference_collectors_layout)
        self._layout.addWidget(self._entity_reference_collectors_widget)

        self._entity_reference_collection_widgets: list[QWidget] = []
        self._entity_reference_remove_buttons: list[QPushButton] = []

        if caption_builder:
            self._caption = QLabel()
            self._layout.addWidget(self._caption)

        self._add_entity_reference_button = QPushButton()
        self._add_entity_reference_button.released.connect(self._add_entity_reference)
        self._layout.addWidget(self._add_entity_reference_button)

        self._build_entity_references_collection()

        self._set_translatables()

    def _add_entity_reference(self) -> None:
        entity_reference = EntityReference['UserFacingEntity & Entity']()
        self._entity_references.append(entity_reference)
        self._build_entity_reference_collection(len(self._entity_references) - 1, entity_reference)
        self._set_translatables()

    def _remove_entity_reference(self, i: int) -> None:
        del self._entity_references[i]
        del self._entity_reference_collectors[i]
        del self._entity_reference_remove_buttons[i]
        self._entity_reference_collectors_layout.removeWidget(self._entity_reference_collection_widgets[i])
        del self._entity_reference_collection_widgets[i]

    def _build_entity_references_collection(self) -> None:
        for i, entity_reference in enumerate(self._entity_references):
            self._build_entity_reference_collection(i, entity_reference)

    def _build_entity_reference_collection(
        self,
        i: int,
        entity_reference: EntityReference[UserFacingEntity & Entity],
    ) -> None:
        widget = QWidget()
        layout = QHBoxLayout()
        widget.setLayout(layout)
        self._entity_reference_collection_widgets.insert(i, widget)
        self._entity_reference_collectors_layout.insertWidget(i, widget)

        entity_reference_collector = EntityReferenceCollector(self._app, entity_reference)
        self._entity_reference_collectors.append(entity_reference_collector)
        layout.addWidget(entity_reference_collector)

        entity_reference_remove_button = QPushButton()
        self._entity_reference_remove_buttons.insert(i, entity_reference_remove_button)
        entity_reference_remove_button.released.connect(lambda: self._remove_entity_reference(i))
        layout.addWidget(entity_reference_remove_button)

    def _do_set_translatables(self) -> None:
        if self._label_builder:
            self._label.setText(self._label_builder())
        if self._caption_builder:
            self._caption.setText(self._caption_builder())
        self._add_entity_reference_button.setText(self._app.localizer._('Add an entity'))
        for entity_reference_remove_button in self._entity_reference_remove_buttons:
            entity_reference_remove_button.setText(self._app.localizer._('Remove'))



from __future__ import annotations

from typing import Any

from PyQt6 import QtGui
from PyQt6.QtCore import QObject
from PyQt6.QtWidgets import QComboBox, QLabel, QWidget, QMainWindow, QMessageBox
from reactives.instance import ReactiveInstance
from reactives.instance.method import reactive_method

from betty.app import App
from betty.gui.text import Caption
from betty.locale import negotiate_locale, get_display_name


class TranslationsLocaleCollector(ReactiveInstance):
    def __init__(self, app: App, allowed_locales: set[str]):
        super().__init__()
        self._app = app
        self._allowed_locales = allowed_locales

        allowed_locale_names: list[tuple[str, str]] = []
        for allowed_locale in allowed_locales:
            locale_name = get_display_name(allowed_locale)
            if locale_name is not None:
                allowed_locale_names.append((
                    allowed_locale,
                    locale_name,
                ))
        allowed_locale_names = sorted(allowed_locale_names, key=lambda x: x[1])

        def _update_configuration_locale() -> None:
            self._app.configuration.locale = self._configuration_locale.currentData()
        self._configuration_locale = QComboBox()
        for i, (locale, locale_name) in enumerate(allowed_locale_names):
            self._configuration_locale.addItem(locale_name, locale)
            if locale == self._app.configuration.locale:
                self._configuration_locale.setCurrentIndex(i)
        self._configuration_locale.currentIndexChanged.connect(_update_configuration_locale)
        self._configuration_locale_label = QLabel()
        self._configuration_locale_caption = Caption()

        self._set_translatables()

    @property
    def locale(self) -> QComboBox:
        return self._configuration_locale

    @property
    def rows(self) -> list[list[Any]]:
        return [
            [self._configuration_locale_label, self._configuration_locale],
            [self._configuration_locale_caption],
        ]

    @reactive_method(on_trigger_call=True)
    def _set_translatables(self) -> None:
        self._configuration_locale_label.setText(self._app.localizer._('Locale'))
        locale = self.locale.currentData()
        if locale:
            translations_locale = negotiate_locale(
                locale,
                set(self._app.localizers.locales),
            )
            if translations_locale is None:
                self._configuration_locale_caption.setText(self._app.localizer._('There are no translations for {locale_name}.').format(
                    locale_name=get_display_name(locale, self._app.localizer.locale),
                ))
            else:
                negotiated_locale_translations_coverage = self._app.localizers.coverage(translations_locale)
                if negotiated_locale_translations_coverage[1] == 0:
                    negotiated_locale_translations_coverage_percentage = 0
                else:
                    negotiated_locale_translations_coverage_percentage = round(100 / (negotiated_locale_translations_coverage[1] / negotiated_locale_translations_coverage[0]))
                self._configuration_locale_caption.setText(self._app.localizer._('The translations for {locale_name} are {coverage_percentage}% complete.').format(
                    locale_name=get_display_name(translations_locale, self._app.localizer.locale),
                    coverage_percentage=round(negotiated_locale_translations_coverage_percentage)
                ))


class _LocalizedObject(ReactiveInstance):
    def __init__(self, app: App, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._app = app

    def showEvent(  # type: ignore[misc]
        self: '_LocalizedObject' & QObject,
        event: QtGui.QShowEvent,
    ) -> None:
        super().showEvent(event)  # type: ignore[misc]
        self._set_translatables()

    @reactive_method(on_trigger_call=True)
    def _set_translatables(self) -> None:
        self._do_set_translatables()

    def _do_set_translatables(self) -> None:
        pass


class LocalizedWidget(  # type: ignore[misc]
    _LocalizedObject,
    QWidget,
):
    pass


class LocalizedMessageBox(  # type: ignore[misc]
    _LocalizedObject,
    QMessageBox,
):
    pass


class LocalizedWindow(  # type: ignore[misc]
    _LocalizedObject,
    QMainWindow,
):
    pass



from __future__ import annotations

import os
import shutil
from contextlib import suppress
from pathlib import Path
from types import TracebackType

from aiofiles.os import link, makedirs


async def link_or_copy(source_path: Path, destination_path: Path) -> None:
    try:
        with suppress(FileExistsError):
            await link(source_path, destination_path)
    except OSError:
        with suppress(shutil.SameFileError):
            shutil.copyfile(source_path, destination_path)


class ChDir:
    def __init__(self, directory_path: Path):
        self._directory_path = directory_path
        self._owd: str | None = None

    async def __aenter__(self) -> None:
        await self.change()

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        await self.revert()

    async def change(self) -> None:
        self._owd = os.getcwd()
        await makedirs(self._directory_path, exist_ok=True)
        os.chdir(self._directory_path)

    async def revert(self) -> None:
        owd = self._owd
        if owd is not None:
            os.chdir(owd)
